"""
 * Copyright(c) 2021 to 2022 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""

import os
import uuid
import inspect
import platform
import ctypes as ct
from ctypes.util import find_library
from functools import wraps
from dataclasses import dataclass
from enum import IntEnum


if 'CYCLONEDDS_PYTHON_NO_IMPORT_LIBS' not in os.environ:
    from .__library__ import library_path, in_wheel


class CycloneDDSLoaderException(Exception):
    pass


def _load(path):
    """Most basic loader, return the loaded DLL on path"""
    try:
        library = ct.CDLL(path)
    except OSError:
        raise CycloneDDSLoaderException(f"Failed to load CycloneDDS library from {path}")
    if not library:
        raise CycloneDDSLoaderException(f"Failed to load CycloneDDS library from {path}")
    return library


def _loader_wheel_gen(rel_path, ext):
    def _loader_wheel():
        if in_wheel:
            return _load(str(library_path))
        return None
    return _loader_wheel


def _loader_cyclonedds_home_gen(name):
    """
        If CYCLONEDDS_HOME is set it is required to be valid and must be used to load.
    """
    def _loader_cyclonedds_home():
        if "CYCLONEDDS_HOME" not in os.environ:
            return None

        return _load(os.path.join(os.environ["CYCLONEDDS_HOME"], name))
    return _loader_cyclonedds_home


def _loader_on_path_gen(name):
    """
        Attempt to load the library without specifying any path at all
        the system might find it with LD_LIBRARY_PATH or the likes.
    """
    def _loader_on_path():
        try:
            lib = find_library("ddsc")
            if lib:
                return _load(lib)
        except CycloneDDSLoaderException:
            pass

        try:
            return _load(name)
        except CycloneDDSLoaderException:
            pass

        return None
    return _loader_on_path



def _loader_install_path():
        try:
            return _load(str(library_path))
        except CycloneDDSLoaderException:
            pass
        return None

_loaders_per_system = {
    "Linux": [
        _loader_wheel_gen(["..", "cyclonedds.libs"], ".so"),
        _loader_cyclonedds_home_gen(f"lib{os.sep}libddsc.so"),
        _loader_on_path_gen("libddsc.so"),
        _loader_install_path
    ],
    "Windows": [
        _loader_wheel_gen(["..", "cyclonedds.libs"], ".dll"),
        _loader_cyclonedds_home_gen(f"bin{os.sep}ddsc.dll"),
        _loader_on_path_gen("ddsc.dll"),
        _loader_install_path
    ],
    "Darwin": [
        _loader_wheel_gen([".dylibs"], ".dylib"),
        _loader_cyclonedds_home_gen(f"lib{os.sep}libddsc.dylib"),
        _loader_on_path_gen("libddsc.dylib"),
        _loader_install_path
    ]
}


def load_cyclonedds() -> ct.CDLL:
    """
        Internal method to load the Cyclone DDS Dynamic Library.
        Handles platform specific naming/configuration.
    """

    if 'CYCLONEDDS_PYTHON_NO_IMPORT_LIBS' in os.environ:
        return

    system = platform.system()
    if system not in _loaders_per_system:
        raise CycloneDDSLoaderException(
            f"You are running on an unknown system configuration {system}, unable to determine the CycloneDDS load path."
        )

    for loader in _loaders_per_system[system]:
        if not loader:
            continue
        lib = loader()
        if lib:
            return lib

    raise CycloneDDSLoaderException(
        "The CycloneDDS library could not be located. "
        "Try setting the CYCLONEDDS_HOME variable to what you used as CMAKE_INSTALL_PREFIX."
    )

def c_call(cname):
    """
        Decorator. Convert a function into call into the class associated dll.
    """

    class DllCall:
        def __init__(self, function):
            self.function = function

        # This gets called when the class is finalized
        def __set_name__(self, cls, name):
            if 'CYCLONEDDS_PYTHON_NO_IMPORT_LIBS' in os.environ:
                return

            s = inspect.signature(self.function)

            # Set c function types based on python type annotations
            cfunc = getattr(cls._dll_handle, cname, None)

            # Sometimes the c function does not exist, unset attr
            if cfunc is None:
                delattr(cls, name)
                return

            # Note: in python 3.10 we get NoneType for voids instead of None
            # This confuses ctypes a lot, so we explicitly test for it
            # We also add the ignore for the error that flake8 generates
            cfunc.restype = s.return_annotation if s.return_annotation != type(None) else None  # noqa: E721

            # Note: ignoring the 'self' argument
            cfunc.argtypes = [p.annotation for i, p in enumerate(s.parameters.values()) if i > 0]

            # Need to rebuild this function to ignore the 'self' attribute
            @wraps(self.function)
            def final_func(self_, *args):
                return cfunc(*args)

            # replace class named method with c call
            setattr(cls, name, final_func)

    return DllCall


def static_c_call(cname):
    """
        Decorator. Convert a function into call into the class associated dll.
    """

    class DllCall:
        def __init__(self, function):
            self.function = function

        # This gets called when the class is finalized
        def __set_name__(self, cls, name):
            if 'CYCLONEDDS_PYTHON_NO_IMPORT_LIBS' in os.environ:
                return

            s = inspect.signature(self.function)

            # Set c function types based on python type annotations
            cfunc = getattr(cls._dll_handle, cname, None)

            # Sometimes the c function does not exist, unset attr
            if cfunc is None:
                delattr(cls, name)
                return

            # Note: in python 3.10 we get NoneType for voids instead of None
            # This confuses ctypes a lot, so we explicitly test for it
            # We also add the ignore for the error that flake8 generates
            cfunc.restype = s.return_annotation if s.return_annotation != type(None) else None  # noqa: E721

            # Note: ignoring the 'self' argument
            cfunc.argtypes = [p.annotation for i, p in enumerate(s.parameters.values()) if i > 0]

            @wraps(self.function)
            def final_func(*args):
                return cfunc(*args)

            # replace class named method with c call
            setattr(cls, name, final_func)

    return DllCall


def c_callable(return_type, argument_types) -> ct.CFUNCTYPE:
    """
        Decorator. Make a C function type based on python type annotations.
    """
    return ct.CFUNCTYPE(return_type, *argument_types)


class DDS:
    """
        Common class for all DDS related classes. This class enables the c_call magic.
    """
    _dll_handle = load_cyclonedds()

    def __init__(self, reference: int) -> None:
        self._ref = reference


@dataclass
class SampleInfo:
    """
    Contains information about the associated data value

    Attributes
    ----------
    sample_state:
        Possible values: :class:`SampleState<cyclonedds.core.SampleState>`
    view_state:
        Possible values: :class:`ViewState<cyclonedds.core.ViewState>`
    instance_state:
        Possible values: :class:`InstanceState<cyclonedds.core.InstanceState>`
    source_timestamp:
        The time (in unix nanoseconds) that the associated sample was written.
    instance_handle:
        Handle to the data instance (if this is a keyed topic)
    """
    sample_state: int
    view_state: int
    instance_state: int
    valid_data: bool
    source_timestamp: int
    instance_handle: int
    publication_handle: int
    disposed_generation_count: int
    no_writers_generation_count: int
    sample_rank: int
    generation_rank: int
    absolute_generation_rank: int


@dataclass
class InvalidSample:
    key_sample: object
    sample_info: SampleInfo


class stat_kind(IntEnum):
    DDS_STAT_KIND_UINT32 = 0
    DDS_STAT_KIND_UINT64 = 1
    DDS_STAT_KIND_LENGTHTIME = 2


class qos_kind(IntEnum):
    DDS_PARTICIPANT_QOS = 0
    DDS_PUBLISHER_QOS = 1
    DDS_SUBSCRIBER_QOS = 2
    DDS_TOPIC_QOS = 3
    DDS_READER_QOS = 4
    DDS_WRITER_QOS = 5


class stat_value(ct.Union):
    _fields_ = [
        ('u32', ct.c_uint32),
        ('u64', ct.c_uint64),
        ('lengthtime', ct.c_uint64)
    ]


class stat_keyvalue(ct.Structure):
    _fields_ = [
        ('name', ct.c_char_p),
        ('kind', ct.c_int),
        ('u', stat_value)
    ]

class dds_c_t:  # noqa N801
    entity = ct.c_int32
    time = ct.c_int64
    duration = ct.c_int64
    instance_handle = ct.c_int64
    domainid = ct.c_uint32
    sample_state = ct.c_int
    view_state = ct.c_int
    instance_state = ct.c_int
    reliability = ct.c_int
    durability = ct.c_int
    history = ct.c_int
    presentation_access_scope = ct.c_int
    type_consistency = ct.c_int
    ingnorelocal = ct.c_int
    ownership = ct.c_int
    liveliness = ct.c_int
    destination_order = ct.c_int
    data_representation_id = ct.c_int16
    qos_p = ct.c_void_p
    qos_provider_p = ct.c_void_p
    qos_kind = ct.c_int32
    attach = ct.c_void_p
    listener_p = ct.c_void_p
    topic_descriptor_p = ct.c_void_p
    returnv = ct.c_int32

    class inconsistent_topic_status(ct.Structure):  # noqa N801
        _fields_ = [('total_count', ct.c_uint32),
                    ('total_count_change', ct.c_int32)]

    class liveliness_lost_status(ct.Structure):  # noqa N801
        _fields_ = [('total_count', ct.c_uint32),
                    ('total_count_change', ct.c_int32)]

    class liveliness_changed_status(ct.Structure):  # noqa N801
        _fields_ = [('alive_count', ct.c_uint32),
                    ('not_alive_count', ct.c_uint32),
                    ('alive_count_change', ct.c_int32),
                    ('not_alive_count_change', ct.c_int32),
                    ('last_publication_handle', ct.c_int64)]

    class offered_deadline_missed_status(ct.Structure):  # noqa N801
        _fields_ = [('total_count', ct.c_uint32),
                    ('total_count_change', ct.c_int32),
                    ('last_instance_handle', ct.c_int64)]

    class offered_incompatible_qos_status(ct.Structure):  # noqa N801
        _fields_ = [('total_count', ct.c_uint32),
                    ('total_count_change', ct.c_int32),
                    ('last_policy_id', ct.c_uint32)]

    class sample_lost_status(ct.Structure):  # noqa N801
        _fields_ = [('total_count', ct.c_uint32),
                    ('total_count_change', ct.c_int32)]

    class sample_rejected_status(ct.Structure):  # noqa N801
        _fields_ = [('total_count', ct.c_uint32),
                    ('total_count_change', ct.c_int32),
                    ('last_reason', ct.c_int),
                    ('last_instance_handle', ct.c_int64)]

    class requested_deadline_missed_status(ct.Structure):  # noqa N801
        _fields_ = [('total_count', ct.c_uint32),
                    ('total_count_change', ct.c_int32),
                    ('last_instance_handle', ct.c_int64)]

    class requested_incompatible_qos_status(ct.Structure):  # noqa N801
        _fields_ = [('total_count', ct.c_uint32),
                    ('total_count_change', ct.c_int32),
                    ('last_policy_id', ct.c_uint32)]

    class publication_matched_status(ct.Structure):  # noqa N801
        _fields_ = [('total_count', ct.c_uint32),
                    ('total_count_change', ct.c_int32),
                    ('current_count', ct.c_uint32),
                    ('current_count_change', ct.c_int32),
                    ('last_subscription_handle', ct.c_int64)]

    class subscription_matched_status(ct.Structure):  # noqa N801
        _fields_ = [('total_count', ct.c_uint32),
                    ('total_count_change', ct.c_int32),
                    ('current_count', ct.c_uint32),
                    ('current_count_change', ct.c_int32),
                    ('last_publication_handle', ct.c_int64)]

    class guid(ct.Structure):  # noqa N801
        _fields_ = [('v', ct.c_uint8 * 16)]

        def as_python_guid(self) -> uuid.UUID:
            return uuid.UUID(bytes=bytes(self.v))

    class sample_info(ct.Structure):  # noqa N801
        _fields_ = [
            ('sample_state', ct.c_uint),
            ('view_state', ct.c_uint),
            ('instance_state', ct.c_uint),
            ('valid_data', ct.c_bool),
            ('source_timestamp', ct.c_int64),
            ('instance_handle', ct.c_uint64),
            ('publication_handle', ct.c_uint64),
            ('disposed_generation_count', ct.c_uint32),
            ('no_writers_generation_count', ct.c_uint32),
            ('sample_rank', ct.c_uint32),
            ('generation_rank', ct.c_uint32),
            ('absolute_generation_rank', ct.c_uint32)
        ]

    class sample_buffer(ct.Structure):  # noqa N801
        _fields_ = [
            ('buf', ct.c_void_p),
            ('len', ct.c_size_t)
        ]

    class statistics(ct.Structure):
        _fields_ = [
            ('entity', ct.c_int32),
            ('opaque', ct.c_uint64),
            ('time', ct.c_int64),
            ('count', ct.c_size_t),
            ('kv', ct.c_void_p)
        ]

    def stat_factory(n_kv: int) -> ct.Structure:
        class vstatistics(ct.Structure):
            _fields_ = [
                ('entity', ct.c_int32),
                ('opaque', ct.c_uint64),
                ('time', ct.c_int64),
                ('count', ct.c_size_t),
                ('kv', stat_keyvalue * n_kv)
            ]
        return vstatistics


try:
    import cyclonedds._clayer as _clayer  # noqa E402
except ImportError as e:
    raise ImportError(f"Error importing Cyclone DDS C library: is Cyclone built using the ENABLE_TYPELIB build option? -- {e}") from e

dds_infinity: int = _clayer.DDS_INFINITY
uint32_max: int = _clayer.UINT32_MAX
feature_typelib = _clayer.HAS_TYPELIB
feature_type_discovery = _clayer.HAS_TYPE_DISCOVERY
feature_topic_discovery = _clayer.HAS_TOPIC_DISCOVERY
dds_domain_default: int = _clayer.DDS_DOMAIN_DEFAULT



"""
 * Copyright(c) 2021 to 2022 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""

from typing import Any, Dict, Union, ForwardRef
from importlib import import_module

from ._type_helper import Annotated, get_origin, get_args
from .types import array, sequence, typedef, case, default, NoneType


class WrapOpt:
    inner: type
    def __init__(self, inner) -> None:
        self.inner = inner

def _strip_unextended_type(module, _type: Any) -> Any:
    if type(_type) == str:
        # Import
        try:
            if '.' in _type:
                # Fully qualified name
                rname, rmodule = _type[::-1].split(".", 1)
                typename, modulename = rname[::-1], rmodule[::-1]
            else:
                # Not fully qualified -> must be same module
                typename, modulename = (_type, module)

            pymodule = import_module(modulename)
            return _strip_unextended_type(modulename, getattr(pymodule, typename))
        except (ImportError, AttributeError):
            raise TypeError(f"Type {_type} as used in {module} cannot be resolved.")
    elif isinstance(_type, (array, sequence, typedef, case, default)):
        _type.subtype = _strip_unextended_type(module, _type.subtype)
        return _type
    if isinstance(_type, ForwardRef):
        return _strip_unextended_type(module, _type.__forward_arg__)

    ori = get_origin(_type)
    if ori is None:
        return _type
    args = get_args(_type)

    if args is None or len(args) == 0:
        return _type
    if ori == Union and NoneType in args and len(args) == 2:
        # optionals
        if args[0] == NoneType:
            return WrapOpt(_strip_unextended_type(module, args[1]))
        else:
            return WrapOpt(_strip_unextended_type(module, args[0]))
    if ori == ForwardRef:
        return _strip_unextended_type(module, args[0])
    if ori == Annotated and len(args) == 2 and type(args[1]) == str:
        # primitive type
        return _type
    if ori == Annotated and len(args) == 2:
        return _strip_unextended_type(module, args[1])

    if len(args) == 1:
        return ori[_strip_unextended_type(module, args[0])]
    elif len(args) == 2:
        return ori[_strip_unextended_type(module, args[0]), _strip_unextended_type(module, args[1])]
    elif len(args) == 3:
        return ori[
            _strip_unextended_type(module, args[0]),
            _strip_unextended_type(module, args[1]),
            _strip_unextended_type(module, args[2])
        ]

    return _type


def _make_extended_type_hints(cls: Any) -> Dict[str, Any]:
    hints = cls.__annotations__
    return {k: _strip_unextended_type(cls.__module__, v) for k, v in hints.items() if not k.startswith("__")}


def _normalize_idl_class(cls: Any) -> None:
    if "__idl_type_annotations__" in cls.__dict__:
        return
    if cls.__name__.startswith("Idl") and cls.__module__ == "cyclonedds.idl":
        return

    cls.__idl_type_annotations__ = _make_extended_type_hints(cls)

    if hasattr(cls, '__idl_discriminator__'):
        cls.__idl_discriminator__ = _strip_unextended_type(cls.__module__, cls.__idl_discriminator__)

    for k in cls.__idl_type_annotations__.keys():
        if k not in cls.__idl_field_annotations__:
            cls.__idl_field_annotations__[k] = {}


def get_extended_type_hints(cls: Any) -> Dict[str, Any]:
    if isinstance(cls, typedef):
        return {'_': _strip_unextended_type(cls.__module__, cls.subtype)}
    _normalize_idl_class(cls)
    return cls.__idl_type_annotations__


def get_idl_field_annotations(cls: Any) -> Dict[str, Dict[str, Any]]:
    return cls.__idl_field_annotations__


def get_idl_annotations(cls: Any) -> Dict[str, Any]:
    return cls.__idl_annotations__


__all__ = ["get_extended_type_hints", "get_idl_field_annotations", "get_idl_annotations"]



"""
 * Copyright(c) 2021 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""

from time import time_ns as _time_ns
from .core import Entity
from .internal import dds_infinity


def isgoodentity(v: object) -> bool:
    """Helper function that checks to see if an object is a valid :class:`Entity<cyclonedds.core.Entity>` returned from DDS.
    This function will never raise an exception.

    Parameters
    ----------
    v : object, optional
        The object to check

    Returns
    -------
    bool
        Whether this entity is a valid :class:`Entity<cyclonedds.core.Entity>`.
    """
    return \
        v is not None and \
        isinstance(v, Entity) and \
        hasattr(v, "_ref") and \
        type(v._ref) == int and \
        v._ref > 0


def duration(*, weeks: float = 0, days: float = 0, hours: float = 0, minutes: float = 0, seconds: float = 0,
             milliseconds: float = 0, microseconds: float = 0, nanoseconds: int = 0, infinite: bool = False) -> int:
    """Durations are always expressed in nanoseconds in DDS (dds_duration_t). This helper function lets
    you write time in a human readable format.

    Examples
    --------
    >>> duration(weeks=2, days=10, minutes=10)

    Parameters
    ----------
        weeks: float, default=0
        days: float, default=0
        hours: float, default=0
        minutes: float, default=0
        seconds: float, default=0
        milliseconds: float, default=0
        microseconds: float, default=0
        nanoseconds: int, default=0
        infinite: bool, default=False

    Returns
    -------
    int
        Duration expressed in nanoseconds.
    """

    if infinite:
        return dds_infinity

    days += weeks * 7
    hours += days * 24
    minutes += hours * 60
    seconds += minutes * 60
    milliseconds += seconds * 1000
    microseconds += milliseconds * 1000
    nanoseconds += microseconds * 1000
    return int(nanoseconds)


class timestamp:
    @staticmethod
    def now():
        """
        In DDS timestamps are typically expressed as nanoseconds since the Unix Epoch (dds_time_t). This helper function
        returns the current time in nanoseconds.

        Returns
        -------
        int
            Number of nanoseconds since the Unix Epoch.
        """
        return _time_ns()



from typing import Type
import pytest

from cyclonedds.domain import DomainParticipant
from cyclonedds.topic import Topic
from cyclonedds.sub import Subscriber, DataReader
from cyclonedds.pub import Publisher, DataWriter
from cyclonedds.core import Qos

from support_modules.testtopics import Message


def test_scoped_qos(common_setup):
    with pytest.raises(TypeError):
        DomainParticipant(qos=Qos().topic())
    DomainParticipant(qos=Qos().domain_participant())

    with pytest.raises(TypeError):
        Topic(common_setup.dp, 'Message', Message, qos=Qos().domain_participant())
    Topic(common_setup.dp, 'Message', Message, qos=Qos().topic())

    with pytest.raises(TypeError):
        Publisher(common_setup.dp, qos=Qos().subscriber())
    Publisher(common_setup.dp, qos=Qos().publisher())

    with pytest.raises(TypeError):
        Subscriber(common_setup.dp, qos=Qos().publisher())
    Subscriber(common_setup.dp, qos=Qos().subscriber())

    with pytest.raises(TypeError):
        DataWriter(common_setup.pub, common_setup.tp, qos=Qos().datareader())
    DataWriter(common_setup.pub, common_setup.tp, qos=Qos().datawriter())

    with pytest.raises(TypeError):
        DataReader(common_setup.sub, common_setup.tp, qos=Qos().datawriter())
    DataReader(common_setup.sub, common_setup.tp, qos=Qos().datareader())


"""
 * Copyright(c) 2021 to 2022 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""

from setuptools.command.build_ext import build_ext as _build_ext
from setuptools import Extension


class Library(Extension):
    pass


class build_ext(_build_ext):
    def get_libraries(self, ext):
        if isinstance(ext, Library):
            return ext.libraries
        return super().get_libraries(ext)

    def get_export_symbols(self, ext):
        if isinstance(ext, Library):
            return ext.export_symbols
        return super().get_export_symbols(ext)




import pytest

import cyclonedds.internal
from cyclonedds.domain import DomainParticipant
from cyclonedds.sub import Subscriber
from cyclonedds.topic import Topic
from cyclonedds.util import duration, isgoodentity
from cyclonedds.builtin import BuiltinDataReader, BuiltinTopicDcpsParticipant, BuiltinTopicDcpsSubscription, BuiltinTopicDcpsTopic

from support_modules.testtopics import Message

requires_dcps_topic = pytest.mark.skipif(
    not cyclonedds.internal.feature_topic_discovery, reason="Cannot use BuiltinTopicDcpsTopic since topic discovery is disabled."
)

def test_builtin_dcps_participant(manual_setup):
    dp = manual_setup.dp
    sub = Subscriber(dp)
    dr1 = BuiltinDataReader(sub, BuiltinTopicDcpsParticipant)
    dr2 = BuiltinDataReader(sub, BuiltinTopicDcpsSubscription)

    assert isgoodentity(dr1)
    assert isgoodentity(dr2)
    assert dr1.take_next().key == dp.guid
    msg = dr2.take(N=2)
    assert {msg[0].key, msg[1].key} == {dr1.guid, dr2.guid}


def test_builtin_dcps_participant_read_next(manual_setup):
    dp = manual_setup.dp
    sub = Subscriber(dp)
    dr1 = BuiltinDataReader(sub, BuiltinTopicDcpsParticipant)
    dr2 = BuiltinDataReader(sub, BuiltinTopicDcpsSubscription)

    assert isgoodentity(dr1)
    assert isgoodentity(dr2)
    assert dr1.read_next().key == dp.guid
    msg = dr2.take(N=2)
    assert {msg[0].key, msg[1].key} == {dr1.guid, dr2.guid}


def test_builtin_dcps_participant_iter(manual_setup):
    dp = manual_setup.dp
    sub = Subscriber(dp)
    dr1 = BuiltinDataReader(sub, BuiltinTopicDcpsParticipant)
    dr2 = BuiltinDataReader(sub, BuiltinTopicDcpsSubscription)

    assert isgoodentity(dr1)
    assert isgoodentity(dr2)

    for msg in dr1.read_iter(timeout=duration(milliseconds=10)):
        assert msg.key == dp.guid

    for msg in dr2.take_iter(timeout=duration(milliseconds=10)):
        assert msg.key in [dr1.guid, dr2.guid]


@requires_dcps_topic
def test_builtin_dcps_topic_read(manual_setup):
    dp = manual_setup.dp
    tdr = BuiltinDataReader(dp, BuiltinTopicDcpsTopic)
    tp = Topic(dp, 'MessageTopic', Message)

    # assert tp.typename == tp.get_type_name() == 'Message'

    assert isgoodentity(tdr)
    assert isgoodentity(tp)

    msg = tdr.read_one(timeout=duration(milliseconds=10))

    assert msg.topic_name == 'MessageTopic'
    assert msg.type_name == 'Message'


@requires_dcps_topic
def test_builtin_dcps_topic_take(manual_setup):
    dp = manual_setup.dp
    tdr = BuiltinDataReader(dp, BuiltinTopicDcpsTopic)

    tp = Topic(dp, 'MessageTopic', Message)

    assert isgoodentity(tdr)
    assert isgoodentity(tp)

    msg = tdr.take_one(timeout=duration(milliseconds=10))

    assert msg.topic_name == 'MessageTopic'
    assert msg.type_name == 'Message'




import pytest

from cyclonedds.domain import DomainParticipant
from cyclonedds.topic import Topic
from cyclonedds.util import isgoodentity

from support_modules.testtopics import Message


def test_create_participant():
    dp = DomainParticipant(0)
    assert isgoodentity(dp)



from cyclonedds.internal import CycloneDDSLoaderException, load_cyclonedds
from cyclonedds.__library__ import library_path, in_wheel
from pytest_mock import MockerFixture
import pytest
import os


if in_wheel:
    pytest.skip("Loader tests are not needed in wheels: loadpath is static.", allow_module_level=True)


def gen_test_loader(loadlist):
    def load(path):
        loadlist.append(path)
        return None
    return load


def gen_joiner(sep):
    def joiner(*args):
        return sep.join(args)
    return joiner


def common_mocks(mocker: MockerFixture, platform: str, ext: str):
    loadlist = []
    mocker.patch("ctypes.util.find_library", new=lambda x: None)
    mocker.patch("cyclonedds.internal._load", new=gen_test_loader(loadlist))
    mocker.patch("platform.system", new=lambda: platform)
    mocker.patch("os.environ", new={"CYCLONEDDS_HOME": "env_canary", "PATH": ""})
    return loadlist


def test_loading_linux(mocker: MockerFixture):
    paths = common_mocks(mocker, "Linux", ".so")
    try:
        load_cyclonedds()
    except CycloneDDSLoaderException:
        pass

    assert paths == [
        f"env_canary{os.sep}lib{os.sep}libddsc.so",
        "libddsc.so",
        library_path
    ]


def test_loading_macos(mocker):
    paths = common_mocks(mocker, "Darwin", ".dylib")
    try:
        load_cyclonedds()
    except CycloneDDSLoaderException:
        pass

    assert paths == [
        f"env_canary{os.sep}lib{os.sep}libddsc.dylib",
        "libddsc.dylib",
        library_path
    ]


def test_loading_windows(mocker):
    paths = common_mocks(mocker, "Windows", ".dll")
    try:
        load_cyclonedds()
    except CycloneDDSLoaderException:
        pass

    assert paths == [
        f"env_canary{os.sep}bin{os.sep}ddsc.dll",
        "ddsc.dll",
        library_path
    ]



"""
 * Copyright(c) 2021 to 2022 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""

from inspect import isclass
from typing import List, Set, Tuple, Type, Optional, Union, Any, Dict
from collections import deque
from dataclasses import dataclass
from hashlib import md5

from ._typesupport.DDS import XTypes as xt
from . import types as pt
from . import annotations as annotate
from . import IdlStruct, IdlUnion, IdlBitmask, IdlEnum, make_idl_struct, make_idl_union, make_idl_enum, make_idl_bitmask
from ._support import Endianness
from ._type_helper import get_origin, get_args
from ._type_normalize import get_extended_type_hints, get_idl_annotations, get_idl_field_annotations, WrapOpt


uint32_max = 2 ** 32 - 1
python_keywords = [
    "False", "None", "True", "and", "as", "assert",
    "break", "class", "continue", "def", "del", "elif",
    "else", "except", "finally", "for", "from", "global",
    "if", "import", "in", "is", "lambda", "nonlocal",
    "not", "or", "pass", "raise", "return", "try",
    "while", "with", "yield", "idl", "annotate", "types",
    "auto", "TYPE_CHECKING", "Optional"]

def _is_optional(_type: Any) -> bool:
    return isinstance(_type, WrapOpt)


def kosaraju(graph: List[List[int]]) -> Tuple[List[int], Dict[int, int], List[Set[int]]]:
    # https://en.wikipedia.org/wiki/Kosaraju%27s_algorithm

    number_of_nodes = len(graph)

    # 1. For each vertex u of the graph, mark u as unvisited. Let L be empty.
    visited = [False] * number_of_nodes
    L = []
    reverse_graph: List[List[int]] = [[] for _ in range(number_of_nodes)]

    def visit(u):
        if not visited[u]:
            visited[u] = True
            for v in graph[u]:
                visit(v)
                reverse_graph[v] += [u]

            # we keep L in normal order instead of prepend
            #  and reverse iterate instead
            L.append(u)

    # 2. For each vertex u of the graph do visit(u)
    for u in range(number_of_nodes):
        visit(u)

    roots = [0] * number_of_nodes
    rootset: Set[int] = set()

    def assign(u, root):
        if visited[u]:
            visited[u] = False
            roots[u] = root
            rootset.add(root)
            for v in reverse_graph[u]:
                assign(v, root)

    # 3: For each element u of l in order, do assign(u, u)
    for u in reversed(L):
        assign(u, u)

    # Busywork: turn root node into component id + component index
    roots_to_components = {r: i for i, r in enumerate(sorted(rootset))}
    component_sizes = {r: 0 for r in range(len(rootset))}
    component_ids = [0] * number_of_nodes
    component_graph: List[Set[int]] = [set() for _ in range(len(rootset))]

    for i in range(number_of_nodes):
        component_id = roots_to_components[roots[i]]
        component_sizes[component_id] += 1
        component_ids[i] = component_id

    for i in range(number_of_nodes):
        c = component_ids[i]
        for u in graph[i]:
            if component_ids[u] != c:
                component_graph[c].add(component_ids[u])

    # component indexes - component sizes - component adjecency graph
    return component_ids, component_sizes, component_graph


class TypeHash:
    def __init__(
        self, *, cls: Type[Any],
        minimal_type_object: Union[xt.TypeObject, xt.MinimalTypeObject, pt.NoneType] = None,
        complete_type_object: Union[xt.TypeObject, xt.CompleteTypeObject, pt.NoneType] = None,
        minimal_type_identifier: Optional[xt.TypeIdentifier] = None,
        complete_type_identifier: Optional[xt.TypeIdentifier] = None,
        minimal_type_object_serialized: Optional[bytes] = None,
        complete_type_object_serialized: Optional[bytes] = None,
        minimal_hash: Optional[xt.TypeObjectHashId] = None,
        complete_hash: Optional[xt.TypeObjectHashId] = None
    ) -> None:
        self.cls: Type[Any] = cls

        if minimal_type_object is not None and complete_type_object is not None:
            self.minimal_type_object: xt.TypeObject = (
                xt.TypeObject(minimal=minimal_type_object) if isinstance(minimal_type_object, xt.MinimalTypeObject)
                else minimal_type_object
            )
            self.complete_type_object: xt.TypeObject = (
                xt.TypeObject(complete=complete_type_object) if isinstance(complete_type_object, xt.CompleteTypeObject)
                else complete_type_object
            )
            self.minimal_type_object_serialized: bytes = (
                minimal_type_object_serialized or self.minimal_type_object.serialize(endianness=Endianness.Little, use_version_2=True)[4:]
            )
            self.complete_type_object_serialized: bytes = (
                complete_type_object_serialized or self.complete_type_object.serialize(endianness=Endianness.Little, use_version_2=True)[4:]
            )

            self.minimal_hash: xt.TypeObjectHashId = minimal_hash or xt.TypeObjectHashId(
                discriminator=xt.EK_MINIMAL,
                value=md5(self.minimal_type_object_serialized).digest()[:14]
            )

            self.complete_hash: xt.TypeObjectHashId = complete_hash or xt.TypeObjectHashId(
                discriminator=xt.EK_COMPLETE,
                value=md5(self.complete_type_object_serialized).digest()[:14]
            )

        self.minimal_type_identifier: xt.TypeIdentifier = minimal_type_identifier or xt.TypeIdentifier(
            discriminator=xt.EK_MINIMAL,
            value=self.minimal_hash.value
        )

        self.complete_type_identifier: xt.TypeIdentifier = complete_type_identifier or xt.TypeIdentifier(
            discriminator=xt.EK_COMPLETE,
            value=self.complete_hash.value
        )


class _StrongComponentResolver:
    def __init__(self, names, types, component_ids, component_size, component_id) -> None:
        self.cnames = [name for i, name in enumerate(names) if component_ids[i] == component_id]
        self.ctypes = [type for i, type in enumerate(types) if component_ids[i] == component_id]

        # 4.b If the Strongly Connected Component (SCC) has multiple types, then sort them
        #     using the lexicographic order of their fully qualified type name
        self.ctypes = [value for i, value in sorted(enumerate(self.ctypes), key=lambda x: self.cnames[x[0]])]
        self.cnames = list(sorted(self.cnames))

        # 4.b.i
        self.minimal_components = [
            xt.StronglyConnectedComponentId(
                sc_component_id=xt.TypeObjectHashId(
                    discriminator=xt.EK_MINIMAL,
                    value=b"\0" * 14
                ),
                scc_length=component_size,
                scc_index=i + 1
            )
            for i, name in enumerate(self.cnames)
        ]

        self.complete_components = [
            xt.StronglyConnectedComponentId(
                sc_component_id=xt.TypeObjectHashId(
                    discriminator=xt.EK_COMPLETE,
                    value=b"\0" * 14
                ),
                scc_length=component_size,
                scc_index=i + 1
            )
            for i, name in enumerate(self.cnames)
        ]

        self.minimal_typemap_by_name = {
            name: self.minimal_components[i]
            for i, name in enumerate(self.cnames)
        }
        self.complete_typemap_by_name = {
            name: self.complete_components[i]
            for i, name in enumerate(self.cnames)
        }

        self.minimal_typemap_by_type = {
            type: self.minimal_components[i]
            for i, type in enumerate(self.ctypes)
            if '.' not in self.cnames[i]
        }
        self.complete_typemap_by_type = {
            type: self.complete_components[i]
            for i, type in enumerate(self.ctypes)
            if '.' not in self.cnames[i]
        }

    def resolve(self, name, type):
        if name in self.minimal_typemap_by_name:
            return self.minimal_typemap_by_name.get(name), self.complete_typemap_by_name.get(name)
        if type in self.minimal_typemap_by_type:
            return self.minimal_typemap_by_type.get(type), self.complete_typemap_by_type.get(type)
        return None, None


class XTBuilder:
    _default_extensibility = "final"
    _type_hash_db: Dict[Union[type, Tuple[str, type]], TypeHash] = {}
    _current_sc_resolver: Optional[_StrongComponentResolver] = None

    @classmethod
    def process_type(cls, topic_type) -> Tuple[xt.TypeInformation, xt.TypeMapping]:
        _names, _types, _graph, no_dependency_types = cls.gather_types(topic_type)

        no_dependency_hashes = [cls._resolve_typehash('', _type) for _type in no_dependency_types]

        if not _graph:
            # Hurray our graph was fully resolved by no_deps types, which means our type is a no_dep type
            hash: TypeHash = cls._resolve_typehash('', topic_type)
            return xt.TypeInformation(
                minimal=xt.TypeIdentifierWithDependencies(
                    typeid_with_size=xt.TypeIdentifierWithSize(
                        type_id=hash.minimal_type_identifier,
                        typeobject_serialized_size=len(hash.minimal_type_object_serialized)
                    ),
                    dependent_typeid_count=0,
                    dependent_typeids=[]
                ),
                complete=xt.TypeIdentifierWithDependencies(
                    typeid_with_size=xt.TypeIdentifierWithSize(
                        type_id=hash.complete_type_identifier,
                        typeobject_serialized_size=len(hash.complete_type_object_serialized)
                    ),
                    dependent_typeid_count=0,
                    dependent_typeids=[]
                )
            ), xt.TypeMapping(
                identifier_object_pair_minimal=[
                    xt.TypeIdentifierTypeObjectPair(
                        type_identifier=hash.minimal_type_identifier,
                        type_object=hash.minimal_type_object
                    )
                ],
                identifier_object_pair_complete=[
                    xt.TypeIdentifierTypeObjectPair(
                        type_identifier=hash.complete_type_identifier,
                        type_object=hash.complete_type_object
                    )
                ],
                identifier_complete_minimal=[
                    xt.TypeIdentifierPair(
                        type_identifier1=hash.complete_type_identifier,
                        type_identifier2=hash.minimal_type_identifier
                    )
                ]
            )

        _type_component_ids, _component_sizes, _component_graph = kosaraju(_graph)

        # Top level type is at position 0 in component 0, guaranteed by gather_types
        assert _names[0] == topic_type.__idl_typename__.replace('.', '::')
        scc_typeids = cls._handle_component(_names, _types, _graph, _type_component_ids, _component_sizes, _component_graph, 0)

        # Construct xt.TypeInformation
        dependent_typeids: List[Tuple[xt.TypeIdentifierWithSize, xt.TypeIdentifierWithSize]] = [
            (
                xt.TypeIdentifierWithSize(
                    type_id=dhash.minimal_type_identifier,
                    typeobject_serialized_size=len(dhash.minimal_type_object_serialized)
                ),
                xt.TypeIdentifierWithSize(
                    type_id=dhash.complete_type_identifier,
                    typeobject_serialized_size=len(dhash.complete_type_object_serialized)
                ),
            )
            for dhash in no_dependency_hashes
        ]
        dependent_typeids += [
            (
                xt.TypeIdentifierWithSize(
                    type_id=dhash.minimal_type_identifier,
                    typeobject_serialized_size=len(dhash.minimal_type_object_serialized)
                ),
                xt.TypeIdentifierWithSize(
                    type_id=dhash.complete_type_identifier,
                    typeobject_serialized_size=len(dhash.complete_type_object_serialized)
                ),
            )
            for dhash in scc_typeids if dhash.cls != topic_type
        ]

        # TODO redo this
        full_list_typehashes: List[TypeHash] = []

        for (name, _type) in zip(_names, _types):
            if _type in cls._type_hash_db:
                full_list_typehashes.append(cls._type_hash_db[_type])
            elif '.' in name and (name, _type) in cls._type_hash_db:
                # member
                full_list_typehashes.append(cls._type_hash_db[(name, _type)])
            else:
                full_list_typehashes.append(cls._resolve_typehash(name, _type))

        for _type in no_dependency_types:
            full_list_typehashes.append(cls._resolve_typehash("", _type))

        hash = full_list_typehashes[0]
        minimal_seen = set()
        minimal_dependend_typeids = [
            d[0] for d in dependent_typeids
            if not (d[0].type_id in minimal_seen or minimal_seen.add(d[0].type_id))
        ]

        return xt.TypeInformation(
            minimal=xt.TypeIdentifierWithDependencies(
                typeid_with_size=xt.TypeIdentifierWithSize(
                    type_id=hash.minimal_type_identifier,
                    typeobject_serialized_size=len(hash.minimal_type_object_serialized)
                ),
                dependent_typeid_count=len(minimal_dependend_typeids),
                dependent_typeids=minimal_dependend_typeids
            ),
            complete=xt.TypeIdentifierWithDependencies(
                typeid_with_size=xt.TypeIdentifierWithSize(
                    type_id=hash.complete_type_identifier,
                    typeobject_serialized_size=len(hash.complete_type_object_serialized)
                ),
                dependent_typeid_count=len(dependent_typeids),
                dependent_typeids=[d[1] for d in dependent_typeids]
            )
        ), xt.TypeMapping(
            identifier_object_pair_minimal=[
                xt.TypeIdentifierTypeObjectPair(
                    type_identifier=dhash.minimal_type_identifier,
                    type_object=dhash.minimal_type_object
                )
                for dhash in full_list_typehashes
            ],
            identifier_object_pair_complete=[
                xt.TypeIdentifierTypeObjectPair(
                    type_identifier=dhash.complete_type_identifier,
                    type_object=dhash.complete_type_object
                )
                for dhash in full_list_typehashes
            ],
            identifier_complete_minimal=[
                xt.TypeIdentifierPair(
                    type_identifier1=dhash.complete_type_identifier,
                    type_identifier2=dhash.minimal_type_identifier
                )
                for dhash in full_list_typehashes
            ]
        )

    @classmethod
    def _deep_gather_type(cls, _type, deep=False):
        # TODO: maps are a problem...

        if isclass(_type) and issubclass(_type, (IdlStruct, IdlUnion, IdlEnum, IdlBitmask)):
            return _type, deep

        if isinstance(_type, pt.sequence) or isinstance(_type, pt.array):
            return cls._deep_gather_type(_type.subtype, False)

        if isinstance(_type, pt.typedef):
            return _type, True

        if isinstance(_type, pt.case) or isinstance(_type, pt.default):
            return cls._deep_gather_type(_type.subtype, False)

        if _is_optional(_type):
            return cls._deep_gather_type(_type.inner, False)

        return None, False

    @classmethod
    def gather_types(cls, _type):
        toscan = deque()
        toscan.append(_type)

        # Start with graph with single node no connections
        graph = {
            _type.__idl_typename__.replace('.', '::'): set()
        }
        graph_types = {
            _type.__idl_typename__.replace('.', '::'): _type
        }

        # 1. Let TypeDependencyDG(T) be the dependency digraph that contains only the types that are reachable from T
        while toscan:
            _ctype = toscan.pop()
            my_node_name = _ctype.__idl_typename__.replace('.', '::')

            if isclass(_ctype) and issubclass(_ctype, IdlUnion):
                # get_extended_type_hints will not inspect the discriminator, and that can be an enum
                discriminator_type = _ctype.__idl_discriminator__
                if isclass(discriminator_type) and issubclass(discriminator_type, IdlEnum):
                    scan_node_name = discriminator_type.__idl_typename__.replace('.', '::')
                    if scan_node_name not in graph:
                        graph[scan_node_name] = set()
                        graph_types[scan_node_name] = discriminator_type
                        toscan.append(discriminator_type)

                    graph[my_node_name].add(scan_node_name)

            if isclass(_ctype) and issubclass(_ctype, IdlStruct):
                # get_extended_type_hints will not inspect the base type, which is a struct
                base_type = _ctype.__base__
                if base_type is None or base_type == IdlStruct:
                    pass
                else:
                    scan_node_name = base_type.__idl_typename__.replace('.', '::')
                    if scan_node_name not in graph:
                        graph[scan_node_name] = set()
                        graph_types[scan_node_name] = base_type
                        toscan.append(base_type)

                    graph[my_node_name].add(scan_node_name)

            for name, fieldtype in get_extended_type_hints(_ctype).items():
                m, deep = cls._deep_gather_type(fieldtype)
                plain = cls._impl_xt_is_plain(fieldtype)

                # m is None if primitive or a type with fields
                if m is not None:
                    if not plain and not isinstance(_ctype, pt.typedef):
                        scan_node_name = m.__idl_typename__.replace('.', '::')
                        depending_name = f"{my_node_name}.{name}"

                        if scan_node_name not in graph:
                            graph[scan_node_name] = set()
                            graph_types[scan_node_name] = m
                            toscan.append(m)

                        graph[depending_name] = set([scan_node_name])
                        graph[my_node_name].add(depending_name)
                        graph_types[depending_name] = fieldtype
                    else:
                        scan_node_name = m.__idl_typename__.replace('.', '::')
                        if scan_node_name not in graph:
                            graph[scan_node_name] = set()
                            graph_types[scan_node_name] = m
                            toscan.append(m)

                        graph[my_node_name].add(scan_node_name)

        # 2. Let ReducedDependencyDG(T) be the subdigraph of TypeDependencyDG(T) where all
        #    the vertices that have no outgoing edges are removed.
        no_dependency_types = []
        for name, edges in graph.copy().items():
            if not edges:
                del graph[name]
                no_dependency_types.append(graph_types[name])
                del graph_types[name]

        # Busywork, need graph made of indexes
        node_names = list(graph.keys())
        node_indexes = {name: i for i, name in enumerate(node_names)}
        node_graph = [[node_indexes[name] for name in graph[nname] if name in node_indexes] for nname in node_names]
        node_types = [graph_types[name] for name in node_names]

        return node_names, node_types, node_graph, no_dependency_types

    @classmethod
    def _resolve_typehash(cls, name: str, _type: Type[Any]) -> TypeHash:
        if _type in cls._type_hash_db:
            return cls._type_hash_db[_type]

        minimal_scc, complete_scc = (
            (None, None) if (cls._current_sc_resolver is None or name == "")
            else cls._current_sc_resolver.resolve(name, _type)
        )

        if minimal_scc is not None and complete_scc is not None:
            return TypeHash(
                cls=_type,
                minimal_type_identifier=xt.TypeIdentifier(sc_component_id=minimal_scc),
                complete_type_identifier=xt.TypeIdentifier(sc_component_id=complete_scc)
            )

        minimal_object = cls._xt_minimal_type_object(_type)
        complete_object = cls._xt_complete_type_object(_type)

        hash = TypeHash(
            cls=_type,
            minimal_type_object=minimal_object,
            complete_type_object=complete_object
        )
        cls._type_hash_db[_type] = hash
        return hash

    @dataclass
    @annotate.final
    class _TypeSeqWrap(IdlStruct):
        seq: pt.sequence[xt.TypeObject]

    @classmethod
    def _handle_component(cls, _names, _types, _graph, _type_component_ids, _component_sizes, _component_graph, component_id):
        # Because the component graph has no cycles depth first is guaranteed to be non-infinite

        scc_typeids = sum([
            cls._handle_component(
                _names, _types, _graph, _type_component_ids, _component_sizes, _component_graph, subcomponent_id
            )
            for subcomponent_id in _component_graph[component_id]
        ], [])

        if _component_sizes[component_id] == 1:
            # 4.a. If the Strongly Connected Component (SCC) has a single type, then use the
            # regular algorithm to compute its xt.TypeIdentifier
            # EXCEPTION: type depends on itself (struct Node {@optional Node n;};)
            for i in range(len(_types)):
                if _type_component_ids[i] == component_id:
                    if i not in _graph[i]:  # Catch the exception of self dependence
                        hash: TypeHash = cls._resolve_typehash(_names[i], _types[i])
                        scc_typeids.append(hash)
                        return scc_typeids
                    break

        cls._current_sc_resolver = _StrongComponentResolver(
            _names,
            _types,
            _type_component_ids,
            _component_sizes[component_id],
            component_id
        )

        minimal_type_objects = [None] * _component_sizes[component_id]
        complete_type_objects = [None] * _component_sizes[component_id]

        # 4.b.ii
        for i, (name, _type) in enumerate(zip(_names, _types)):
            if _type_component_ids[i] == component_id:
                minimal_type_objects[cls._current_sc_resolver.cnames.index(name)] = xt.TypeObject(minimal=cls._xt_minimal_type_object(_type))
                complete_type_objects[cls._current_sc_resolver.cnames.index(name)] = xt.TypeObject(complete=cls._xt_complete_type_object(_type))

        # 4.c, 4.d
        # Note: the [4:] is to skip the XCDR header
        minimal_bytes = cls._TypeSeqWrap(seq=minimal_type_objects).serialize(endianness=Endianness.Little, use_version_2=True)[4:]
        complete_bytes = cls._TypeSeqWrap(seq=complete_type_objects).serialize(endianness=Endianness.Little, use_version_2=True)[4:]

        # 4.e
        minimal_hash = xt.TypeObjectHashId(
            discriminator=xt.EK_MINIMAL,
            value=md5(minimal_bytes).digest()[:14]
        )
        complete_hash = xt.TypeObjectHashId(
            discriminator=xt.EK_COMPLETE,
            value=md5(complete_bytes).digest()[:14]
        )

        # 4.f
        for i, (name, _type) in enumerate(zip(_names, _types)):
            if _type_component_ids[i] == component_id:
                if '.' in name:
                    key = (name, _type)
                else:
                    key = _type
                index = cls._current_sc_resolver.cnames.index(name)
                cls._type_hash_db[key] = TypeHash(
                    cls=_type,
                    minimal_type_object=minimal_type_objects[index],
                    complete_type_object=complete_type_objects[index],
                    minimal_type_identifier=xt.TypeIdentifier(
                        sc_component_id=xt.StronglyConnectedComponentId(
                            sc_component_id=minimal_hash,
                            scc_length=_component_sizes[component_id],
                            scc_index=index + 1
                        )
                    ),
                    complete_type_identifier=xt.TypeIdentifier(
                        sc_component_id=xt.StronglyConnectedComponentId(
                            sc_component_id=complete_hash,
                            scc_length=_component_sizes[component_id],
                            scc_index=index + 1
                        )
                    )
                )

        for comp in cls._current_sc_resolver.complete_components:
            comp.sc_component_id = complete_hash

        for comp in cls._current_sc_resolver.minimal_components:
            comp.sc_component_id = minimal_hash

        cls._current_sc_resolver = None

        # Index 0 is used to indicate the full StronglyConnectedComponent
        scc_typeids.append(TypeHash(
            cls=None,
            minimal_type_object=minimal_type_objects,
            minimal_type_object_serialized=minimal_bytes,
            minimal_hash=minimal_hash,
            minimal_type_identifier=xt.TypeIdentifier(
                sc_component_id=xt.StronglyConnectedComponentId(
                    sc_component_id=minimal_hash,
                    scc_length=_component_sizes[component_id],
                    scc_index=0
                )
            ),
            complete_type_object=complete_type_objects,
            complete_type_object_serialized=complete_bytes,
            complete_hash=complete_hash,
            complete_type_identifier=xt.TypeIdentifier(
                sc_component_id=xt.StronglyConnectedComponentId(
                    sc_component_id=complete_hash,
                    scc_length=_component_sizes[component_id],
                    scc_index=0
                )
            )
        ))
        return scc_typeids

    @classmethod
    def _impl_xt_is_fully_descriptive(cls, _type: type) -> bool:
        """Check if a type is fully descriptive, e.g. does not contain any type hashes"""
        if _type in pt._type_code_align_size_default_mapping:
            return True
        if _type in [str, pt.char]:
            return True
        if isinstance(_type, pt.bounded_str):
            return True
        if isinstance(_type, pt.array) or isinstance(_type, pt.sequence):
            return cls._impl_xt_is_fully_descriptive(_type.subtype)
        if get_origin(_type) == Dict:
            args = get_args(_type)
            return cls._impl_xt_is_fully_descriptive(args[0]) and cls._impl_xt_is_fully_descriptive(args[1])
        if _is_optional(_type):
            return cls._impl_xt_is_fully_descriptive(_type.inner)

        return False

    @classmethod
    def _impl_xt_is_plain(cls, _type: type) -> bool:
        """Check if a type is fully descriptive, e.g. does not contain any type hashes"""
        if _type in pt._type_code_align_size_default_mapping:
            return True
        if _type in [str, pt.char]:
            return True
        if isinstance(_type, pt.bounded_str):
            return True
        if isinstance(_type, pt.array):
            inner = _type
            while isinstance(inner, pt.array):
                inner = inner.subtype

            return cls._impl_xt_is_plain(inner)
        if isinstance(_type, pt.sequence):
            return cls._impl_xt_is_plain(_type.subtype)
        if get_origin(_type) == Dict:
            args = get_args(_type)
            return cls._impl_xt_is_plain(args[0]) and cls._impl_xt_is_plain(args[1])
        if isclass(_type) and issubclass(_type, (IdlStruct, IdlUnion, IdlEnum, IdlBitmask)):
            return True
        if isinstance(_type, pt.typedef):
            return True
        if isinstance(_type, (pt.case, pt.default)):
            return cls._impl_xt_is_plain(_type.subtype)
        if isinstance(_type, WrapOpt):
            return cls._impl_xt_is_plain(_type.inner)

        return False

    @classmethod
    def _xt_type_identifier(cls, name: str, entity: Any, minimal: bool) -> xt.TypeIdentifier:
        if entity is None or entity == object:
            return xt.TypeIdentifier(discriminator=xt.TK_NONE, value=None)

        if _is_optional(entity):
            # turned into annotation one level up
            entity = entity.inner

        if entity == pt.char:
            return xt.TypeIdentifier(discriminator=xt.TK_CHAR8, value=None)

        if entity in pt._type_code_align_size_default_mapping:
            return {
                bool: xt.TypeIdentifier(discriminator=xt.TK_BOOLEAN, value=None),
                float: xt.TypeIdentifier(discriminator=xt.TK_FLOAT64, value=None),
                int: xt.TypeIdentifier(discriminator=xt.TK_INT64, value=None),
                pt.int8: xt.TypeIdentifier(discriminator=xt.TK_INT8, value=None),
                pt.int16: xt.TypeIdentifier(discriminator=xt.TK_INT16, value=None),
                pt.int32: xt.TypeIdentifier(discriminator=xt.TK_INT32, value=None),
                pt.int64: xt.TypeIdentifier(discriminator=xt.TK_INT64, value=None),
                pt.byte: xt.TypeIdentifier(discriminator=xt.TK_BYTE, value=None),
                pt.uint8: xt.TypeIdentifier(discriminator=xt.TK_UINT8, value=None),
                pt.uint16: xt.TypeIdentifier(discriminator=xt.TK_UINT16, value=None),
                pt.uint32: xt.TypeIdentifier(discriminator=xt.TK_UINT32, value=None),
                pt.uint64: xt.TypeIdentifier(discriminator=xt.TK_UINT64, value=None),
                pt.char: xt.TypeIdentifier(discriminator=xt.TK_CHAR8, value=None),
                pt.wchar: xt.TypeIdentifier(discriminator=xt.TK_CHAR16, value=None),
                pt.float32: xt.TypeIdentifier(discriminator=xt.TK_FLOAT32, value=None),
                pt.float64: xt.TypeIdentifier(discriminator=xt.TK_FLOAT64, value=None)
            }[entity]

        if cls._impl_xt_is_plain(entity):
            if entity is str:
                return xt.TypeIdentifier(string_sdefn=xt.StringSTypeDefn(bound=0))
            elif isinstance(entity, pt.bounded_str):
                if entity.max_length <= 255:
                    return xt.TypeIdentifier(string_sdefn=xt.StringSTypeDefn(bound=entity.max_length))
                else:
                    return xt.TypeIdentifier(string_ldefn=xt.StringLTypeDefn(bound=entity.max_length))
            elif isinstance(entity, pt.sequence):
                fully_descriptive = cls._impl_xt_is_fully_descriptive(entity)
                header = xt.PlainCollectionHeader(
                    equiv_kind=xt.EK_BOTH if fully_descriptive else (xt.EK_MINIMAL if minimal else xt.EK_COMPLETE),
                    element_flags=xt.MemberFlag(TRY_CONSTRUCT1=True)  # TODO: elements can be external, try_construct
                )

                if entity.max_length is None:
                    return xt.TypeIdentifier(
                        seq_sdefn=xt.PlainSequenceSElemDefn(
                            header=header,
                            bound=0,
                            element_identifier=cls._xt_type_identifier(name, entity.subtype, minimal)
                        )
                    )
                elif entity.max_length <= 255:
                    return xt.TypeIdentifier(
                        seq_sdefn=xt.PlainSequenceSElemDefn(
                            header=header,
                            bound=entity.max_length,
                            element_identifier=cls._xt_type_identifier(name, entity.subtype, minimal)
                        )
                    )
                else:
                    return xt.TypeIdentifier(
                        seq_ldefn=xt.PlainSequenceLElemDefn(
                            header=header,
                            bound=entity.max_length or uint32_max,
                            element_identifier=cls._xt_type_identifier(name, entity.subtype, minimal)
                        )
                    )

            elif isinstance(entity, pt.array):
                fully_descriptive = cls._impl_xt_is_fully_descriptive(entity)
                header = xt.PlainCollectionHeader(
                    equiv_kind=xt.EK_BOTH if fully_descriptive else (xt.EK_MINIMAL if minimal else xt.EK_COMPLETE),
                    element_flags=xt.MemberFlag(TRY_CONSTRUCT1=True)  # TODO: elements can be external, try_construct
                )

                inner: Type[Any] = entity
                bounds: List[int] = []
                while isinstance(inner, pt.array):
                    bounds.append(inner.length)
                    inner = inner.subtype

                if all(b < 256 for b in bounds):
                    return xt.TypeIdentifier(
                        array_sdefn=xt.PlainArraySElemDefn(
                            header=header,
                            array_bound_seq=bounds,
                            element_identifier=cls._xt_type_identifier(name, inner, minimal)
                        )
                    )
                else:
                    return xt.TypeIdentifier(
                        array_ldefn=xt.PlainArrayLElemDefn(
                            header=header,
                            array_bound_seq=bounds,
                            element_identifier=cls._xt_type_identifier(name, inner, minimal)
                        )
                    )
            elif get_origin(entity) == Dict:
                fully_descriptive = cls._impl_xt_is_fully_descriptive(entity)
                keytype, valuetype = get_args(entity)
                header = xt.PlainCollectionHeader(
                    equiv_kind=xt.EK_BOTH if fully_descriptive else (xt.EK_MINIMAL if minimal else xt.EK_COMPLETE),
                    element_flags=xt.MemberFlag()  # TODO: elements can be external, try_construct
                )
                # No small maps
                return xt.TypeIdentifier(
                    map_ldefn=xt.PlainMapLTypeDefn(
                        header=header,
                        bound=uint32_max,
                        element_identifier=cls._xt_type_identifier(name, valuetype, minimal),
                        key_flags=xt.MemberFlag(),  # TODO: keys can be external, try_construct
                        key_identifier=cls._xt_type_identifier(name, keytype, minimal)
                    )
                )

        typehash = cls._resolve_typehash(name, entity)
        if minimal:
            return typehash.minimal_type_identifier
        else:
            return typehash.complete_type_identifier

    @classmethod
    def _xt_minimal_type_object(cls, entity: Any) -> xt.MinimalTypeObject:
        if _is_optional(entity):
            return cls._xt_minimal_type_object(entity.inner)

        if isinstance(entity, pt.typedef):
            return xt.MinimalTypeObject(alias_type=cls._xt_minimal_alias_type(entity))
        elif False:
            # python has no annotation types yet
            return xt.MinimalTypeObject(annotation_type=None)
        elif isclass(entity) and issubclass(entity, IdlStruct):
            entity.__idl__.populate()
            return xt.MinimalTypeObject(struct_type=cls._xt_minimal_struct_type(entity))
        elif isclass(entity) and issubclass(entity, IdlUnion):
            entity.__idl__.populate()
            return xt.MinimalTypeObject(union_type=cls._xt_minimal_union_type(entity))
        elif isinstance(entity, pt.sequence):
            return xt.MinimalTypeObject(sequence_type=cls._xt_minimal_sequence_type(entity))
        elif isinstance(entity, pt.array):
            return xt.MinimalTypeObject(array_type=cls._xt_minimal_array_type(entity))
        elif isinstance(entity, Dict):
            return xt.MinimalTypeObject(map_type=cls._xt_minimal_map_type(entity))
        elif isclass(entity) and issubclass(entity, IdlEnum):
            return xt.MinimalTypeObject(enumerated_type=cls._xt_minimal_enumerated_type(entity))
        elif isclass(entity) and issubclass(entity, IdlBitmask):
            return xt.MinimalTypeObject(bitmask_type=cls._xt_minimal_bitmask_type(entity))
        raise Exception(f"Cannot construct typeobject for entity {entity}")

    @classmethod
    def _xt_complete_type_object(cls, entity: Any) -> xt.CompleteTypeObject:
        if _is_optional(entity):
            return cls._xt_complete_type_object(entity.inner)

        if isinstance(entity, pt.typedef):
            return xt.CompleteTypeObject(alias_type=cls._xt_complete_alias_type(entity))
        elif False:
            # python has no annotation types yet
            return xt.CompleteTypeObject(annotation_type=None)
        elif isclass(entity) and issubclass(entity, IdlStruct):
            entity.__idl__.populate()
            return xt.CompleteTypeObject(struct_type=cls._xt_complete_struct_type(entity))
        elif isclass(entity) and issubclass(entity, IdlUnion):
            entity.__idl__.populate()
            return xt.CompleteTypeObject(union_type=cls._xt_complete_union_type(entity))
        elif isinstance(entity, pt.sequence):
            return xt.CompleteTypeObject(sequence_type=cls._xt_complete_sequence_type(entity))
        elif isinstance(entity, pt.array):
            return xt.CompleteTypeObject(array_type=cls._xt_complete_array_type(entity))
        elif isinstance(entity, Dict):
            return xt.CompleteTypeObject(map_type=cls._xt_complete_map_type(entity))
        elif isclass(entity) and issubclass(entity, IdlEnum):
            return xt.CompleteTypeObject(enumerated_type=cls._xt_complete_enumerated_type(entity))
        elif isclass(entity) and issubclass(entity, IdlBitmask):
            return xt.CompleteTypeObject(bitmask_type=cls._xt_complete_bitmask_type(entity))
        raise Exception(f"Cannot construct typeobject for entity {entity}")

    @classmethod
    def _xt_minimal_alias_type(cls, entity: Any) -> xt.MinimalAliasType:
        return xt.MinimalAliasType(
            alias_flags=xt.TypeFlag(),  # unused, no flags apply
            header=xt.MinimalAliasHeader(),  # empty...
            body=xt.MinimalAliasBody(
                common=xt.CommonAliasBody(
                    related_flags=xt.MemberFlag(),  # unused, no flags apply
                    related_type=cls._xt_type_identifier('', entity.subtype, True)
                )
            )
        )

    @classmethod
    def _xt_complete_alias_type(cls, entity: Any) -> xt.CompleteAliasType:
        return xt.CompleteAliasType(
            alias_flags=xt.TypeFlag(),  # unused, no flags apply
            header=xt.CompleteAliasHeader(
                detail=xt.CompleteTypeDetail(
                    ann_builtin=None,
                    ann_custom=None,
                    type_name=entity.__idl_typename__.replace('.', '::')
                )
            ),
            body=xt.CompleteAliasBody(
                common=xt.CommonAliasBody(
                    related_flags=xt.MemberFlag(),  # unused, no flags apply
                    related_type=cls._xt_type_identifier('', entity.subtype, False)
                ),
                ann_builtin=None,
                ann_custom=None
            )
        )

    @classmethod
    def _xt_minimal_struct_type(cls, entity: Type[IdlStruct]) -> xt.MinimalStructType:
        return xt.MinimalStructType(
            struct_flags=cls._xt_type_flag(entity),
            header=cls._xt_minimal_struct_header(entity),
            member_seq=cls._xt_minimal_struct_member_seq(entity)
        )

    @classmethod
    def _xt_complete_struct_type(cls, entity: Type[IdlStruct]) -> xt.CompleteStructType:
        return xt.CompleteStructType(
            struct_flags=cls._xt_type_flag(entity),
            header=cls._xt_complete_struct_header(entity),
            member_seq=cls._xt_complete_struct_member_seq(entity)
        )

    @classmethod
    def _xt_type_flag(cls, entity: Any) -> xt.TypeFlag:
        flag = xt.TypeFlag()
        cls_a = get_idl_annotations(entity)

        extensibility = cls_a.get("extensibility", cls._default_extensibility)

        if extensibility == "final":
            flag.IS_FINAL = True
        elif extensibility == "appendable":
            flag.IS_APPENDABLE = True
        elif extensibility == "mutable":
            flag.IS_MUTABLE = True

        if cls_a.get("nested", False):
            flag.IS_NESTED = True

        if cls_a.get("autoid", "sequential") == "hash":
            flag.IS_AUTOID_HASH = True

        return flag

    @classmethod
    def _xt_minimal_struct_header(cls, entity: Type[IdlStruct]) -> xt.MinimalStructHeader:
        if entity.__base__ is None or entity.__base__ == IdlStruct:
            return xt.MinimalStructHeader(
                base_type=xt.TypeIdentifier(discriminator=xt.TK_NONE, value=None),
                detail=cls._xt_minimal_type_detail(entity)
            )

        return xt.MinimalStructHeader(
            base_type=cls._xt_type_identifier("_base_", entity.__base__, True),
            detail=cls._xt_minimal_type_detail(entity)
        )

    @classmethod
    def _xt_complete_struct_header(cls, entity: Type[IdlStruct]) -> xt.CompleteStructHeader:
        if entity.__base__ is None or entity.__base__ == IdlStruct:
            return xt.CompleteStructHeader(
                base_type=xt.TypeIdentifier(discriminator=xt.TK_NONE, value=None),
                detail=cls._xt_complete_type_detail(entity)
            )

        return xt.CompleteStructHeader(
            base_type=cls._xt_type_identifier("_base_", entity.__base__, False),
            detail=cls._xt_complete_type_detail(entity)
        )

    @classmethod
    def _xt_minimal_type_detail(cls, entity: Any) -> xt.MinimalTypeDetail:
        # xtypes spec: left for future expansion
        return xt.MinimalTypeDetail()

    @classmethod
    def _xt_complete_type_detail(cls, entity: Any) -> xt.CompleteTypeDetail:
        return xt.CompleteTypeDetail(
            ann_builtin=None,
            ann_custom=None,
            type_name=entity.__idl_typename__.replace(".", "::") if hasattr(entity, "__idl_typename__") else "anonymous"
        )

    @classmethod
    def _xt_minimal_struct_member_seq(cls, entity: Type[IdlStruct]) -> xt.MinimalStructMemberSeq:
        parent_fields = set(
            get_extended_type_hints(entity.__base__).keys()
        ) if entity.__base__ not in (IdlStruct, None) else set()

        return [
            cls._xt_minimal_struct_member(entity, name, _type)
            for name, _type in get_extended_type_hints(entity).items()
            if name not in parent_fields
        ]

    @classmethod
    def _xt_complete_struct_member_seq(cls, entity: Type[IdlStruct]) -> xt.CompleteStructMemberSeq:
        parent_fields = set(
            get_extended_type_hints(entity.__base__).keys()
        ) if entity.__base__ not in (IdlStruct, None) else set()

        return [
            cls._xt_complete_struct_member(entity, name, _type)
            for name, _type in get_extended_type_hints(entity).items()
            if name not in parent_fields
        ]

    @classmethod
    def _xt_minimal_struct_member(cls, entity: Type[IdlStruct], name: str, _type: type) -> xt.MinimalStructMember:
        return xt.MinimalStructMember(
            common=cls._xt_common_struct_member(entity, name, _type, True),
            detail=cls._xt_minimal_member_detail(entity, name, _type)
        )

    @classmethod
    def _xt_complete_struct_member(cls, entity: Type[IdlStruct], name: str, _type: type) -> xt.CompleteStructMember:
        # Reserved Python keywords support (#105)
        annotations = get_idl_field_annotations(entity).get(name)
        realName = annotations["name"] if annotations and "name" in annotations else name
        #######################

        return xt.CompleteStructMember(
            common=cls._xt_common_struct_member(entity, name, _type, False),
            #Reserved Python keywords support (Issue 105):
            detail=cls._xt_complete_member_detail(entity, realName, _type)
            #######################
        )

    @classmethod
    def _xt_common_struct_member(cls, entity: Type[IdlStruct], name: str, _type: type, minimal: bool) -> xt.CommonStructMember:
        return xt.CommonStructMember(
            member_id=entity.__idl__.get_member_id(name),
            member_flags=cls._xt_member_flag(entity, name, _type),
            member_type_id=cls._xt_type_identifier(f"{entity.__idl_typename__.replace('.','::')}.{name}", _type, minimal)
        )

    @classmethod
    def _xt_member_flag(cls, entity: Any, name: str, _type: type) -> xt.MemberFlag:
        flag = xt.MemberFlag()
        field_a = get_idl_field_annotations(entity)
        annotations = field_a.get(name, {})
        in_keylist = name in get_idl_annotations(entity).get("keylist", [])

        # TRY CONSTRUCT TODO: INVALID, DISCARD, TRIM?
        if "default" in annotations:
            # 10 = use_default
            flag.TRY_CONSTRUCT2 = True
        else:
            # 01 = discard
            flag.TRY_CONSTRUCT1 = True

        if annotations.get("external", False):
            flag.IS_EXTERNAL = True

        if _is_optional(_type):
            flag.IS_OPTIONAL = True

        if annotations.get("must_understand", False):
            flag.IS_MUST_UNDERSTAND = True

        if annotations.get("key", False) or in_keylist:
            flag.IS_KEY = True

        if isinstance(_type, pt.default):
            flag.IS_DEFAULT = True

        return flag

    @classmethod
    def _xt_minimal_member_detail(cls, entity: Any, name: str, _type: type) -> xt.MinimalMemberDetail:
        hash = md5(name.encode('utf8')).digest()
        return xt.MinimalMemberDetail(
            name_hash=hash[0:4]
        )

    @classmethod
    def _xt_complete_member_detail(cls, entity: Any, name: str, _type: type) -> xt.CompleteMemberDetail:
        field_a = get_idl_field_annotations(entity)
        annotations = field_a.get(name, {})
        return xt.CompleteMemberDetail(
            name=name,
            ann_builtin=cls._xt_applied_builtin_member_annotations(annotations, _type),
            ann_custom=None
        )

    @classmethod
    def _xt_applied_builtin_member_annotations(
        cls, annotations: Dict[str, Any], _type: Any
    ) -> Optional[xt.AppliedBuiltinMemberAnnotations]:
        if "hash_id" in annotations:
            return xt.AppliedBuiltinMemberAnnotations(
                unit=None,  # unused so far
                min=None,  # unused so far
                max=None,  # unused so far
                hash_id=annotations["hash_id"] or ""
            )
        return None

    @classmethod
    def _xt_minimal_union_type(cls, entity: Type[IdlUnion]) -> xt.MinimalUnionType:
        return xt.MinimalUnionType(
            union_flags=cls._xt_type_flag(entity),
            header=cls._xt_minimal_union_header(entity),
            discriminator=cls._xt_minimal_discriminator_member(entity),
            member_seq=cls._xt_minimal_union_member_seq(entity)
        )

    @classmethod
    def _xt_complete_union_type(cls, entity: Type[IdlUnion]) -> xt.CompleteUnionType:
        return xt.CompleteUnionType(
            union_flags=cls._xt_type_flag(entity),
            header=cls._xt_complete_union_header(entity),
            discriminator=cls._xt_complete_discriminator_member(entity),
            member_seq=cls._xt_complete_union_member_seq(entity)
        )

    @classmethod
    def _xt_minimal_union_header(cls, entity: Type[IdlUnion]) -> xt.MinimalUnionHeader:
        return xt.MinimalUnionHeader(
            detail=cls._xt_minimal_type_detail(entity)
        )

    @classmethod
    def _xt_complete_union_header(cls, entity: Type[IdlUnion]) -> xt.CompleteUnionHeader:
        return xt.CompleteUnionHeader(
            detail=cls._xt_complete_type_detail(entity)
        )

    @classmethod
    def _xt_minimal_discriminator_member(cls, entity: Type[IdlUnion]) -> xt.MinimalDiscriminatorMember:
        return xt.MinimalDiscriminatorMember(
            common=cls._xt_common_discriminator_member(entity, True)
        )

    @classmethod
    def _xt_complete_discriminator_member(cls, entity: Type[IdlUnion]) -> xt.CompleteDiscriminatorMember:
        return xt.CompleteDiscriminatorMember(
            common=cls._xt_common_discriminator_member(entity, False),
            ann_builtin=None,
            ann_custom=None,
        )

    @classmethod
    def _xt_common_discriminator_member(cls, entity: Type[IdlUnion], minimal: bool) -> xt.CommonDiscriminatorMember:
        return xt.CommonDiscriminatorMember(
            member_flags=xt.MemberFlag(IS_KEY=entity.__idl_discriminator_is_key__, IS_MUST_UNDERSTAND=True, TRY_CONSTRUCT1=True),
            type_id=cls._xt_type_identifier("discriminator", entity.__idl_discriminator__, minimal)
        )

    @classmethod
    def _xt_minimal_union_member_seq(cls, entity: Type[IdlUnion]) -> xt.MinimalUnionMemberSeq:
        parent_fields = set(
            get_extended_type_hints(entity.__base__).keys()
        ) if entity.__base__ not in (IdlUnion, None) else set()

        return [
            cls._xt_minimal_union_member(entity, name, _type)
            for name, _type in get_extended_type_hints(entity).items()
            if name not in parent_fields
        ]

    @classmethod
    def _xt_complete_union_member_seq(cls, entity: Type[IdlUnion]) -> xt.CompleteUnionMemberSeq:
        parent_fields = set(
            get_extended_type_hints(entity.__base__).keys()
        ) if entity.__base__ not in (IdlUnion, None) else set()

        return [
            cls._xt_complete_union_member(entity, name, _type)
            for name, _type in get_extended_type_hints(entity).items()
            if name not in parent_fields
        ]

    @classmethod
    def _xt_minimal_union_member(
        cls, entity: Type[IdlUnion], name: str, _type: Type[Any]
    ) -> xt.MinimalUnionMember:
        return xt.MinimalUnionMember(
            common=cls._xt_common_union_member(entity, name, _type, True),
            detail=cls._xt_minimal_member_detail(entity, name, _type.subtype)
        )

    @classmethod
    def _xt_complete_union_member(
        cls, entity: Type[IdlUnion], name: str, _type: Type[Any]
    ) -> xt.CompleteUnionMember:
        # Reserved Python keywords support (#105)
        annotations = get_idl_field_annotations(entity).get(name)
        realName = annotations["name"] if annotations and "name" in annotations else name
        #######################

        return xt.CompleteUnionMember(
            common=cls._xt_common_union_member(entity, name, _type, False),
            #Reserved Python keywords support (Issue 105):
            detail=cls._xt_complete_member_detail(entity, realName, _type.subtype)
            #######################
        )

    @classmethod
    def _xt_common_union_member(
        cls, entity: Type[IdlUnion], name: str, _type: Type[Any], minimal: bool
    ) -> xt.CommonUnionMember:
        return xt.CommonUnionMember(
            member_id=entity.__idl__.get_member_id(name),
            member_flags=cls._xt_member_flag(entity, name, _type),
            type_id=cls._xt_type_identifier(f"{entity.__idl_typename__.replace('.','::')}.{name}", _type.subtype, minimal),
            label_seq=cls._xt_union_case_label_seq(entity, name, _type)
        )

    @classmethod
    def _xt_union_case_label_seq(cls, entity: Type[IdlUnion], name: str, _type: Type[Any]) -> xt.UnionCaseLabelSeq:
        if isclass(entity.__idl_discriminator__) and issubclass(entity.__idl_discriminator__, IdlEnum):
            return [l.value for l in _type.labels] if isinstance(_type, pt.case) else []
        return _type.labels if isinstance(_type, pt.case) else []

    @classmethod
    def _xt_minimal_sequence_type(cls, seq: pt.sequence) -> xt.MinimalSequenceType:
        return xt.MinimalSequenceType(
            collection_flag=xt.TypeFlag(),  # Unused, no flags apply
            header=cls._xt_minimal_collection_header(seq),
            element=cls._xt_minimal_collection_element(seq.subtype, xt.TypeFlag(TRY_CONSTRUCT1=True))
        )

    @classmethod
    def _xt_complete_sequence_type(cls, seq: pt.sequence) -> xt.CompleteSequenceType:
        return xt.CompleteSequenceType(
            collection_flag=xt.TypeFlag(),  # Unused, no flags apply
            header=cls._xt_complete_collection_header(seq),
            element=cls._xt_complete_collection_element(seq.subtype, xt.TypeFlag(TRY_CONSTRUCT1=True))
        )

    @classmethod
    def _xt_minimal_collection_header(cls, collection: Union[pt.array, pt.sequence]) -> xt.MinimalCollectionHeader:
        return xt.MinimalCollectionHeader(
            common=cls._xt_common_collection_header(collection)
        )

    @classmethod
    def _xt_complete_collection_header(cls, collection: Union[pt.array, pt.sequence]) -> xt.CompleteCollectionHeader:
        return xt.CompleteCollectionHeader(
            common=cls._xt_common_collection_header(collection),
            detail=cls._xt_complete_type_detail(collection)
        )

    @classmethod
    def _xt_common_collection_header(cls, collection: Union[pt.array, pt.sequence]) -> xt.CommonCollectionHeader:
        return xt.CommonCollectionHeader(
            bound=uint32_max if isinstance(collection, pt.array) or collection.max_length is None else collection.max_length
        )

    @classmethod
    def _xt_minimal_collection_element(cls, _type: Any, flags: xt.TypeFlag) -> xt.MinimalCollectionElement:
        return xt.MinimalCollectionElement(
            common=cls._xt_common_collection_element(_type, True, flags)
        )

    @classmethod
    def _xt_complete_collection_element(cls, _type: Any, flags: xt.TypeFlag) -> xt.CompleteCollectionElement:
        return xt.CompleteCollectionElement(
            common=cls._xt_common_collection_element(_type, True, flags),
            detail=cls._xt_complete_element_detail(_type)
        )

    @classmethod
    def _xt_complete_element_detail(cls, _type: Any) -> xt.CompleteElementDetail:
        return xt.CompleteElementDetail(
            ann_builtin=cls._xt_applied_builtin_member_annotations({}, _type),
            ann_custom=None
        )

    @classmethod
    def _xt_common_collection_element(cls, _type: Any, minimal: bool, flags: xt.TypeFlag) -> xt.CommonCollectionElement:
        # TODO: how to scope this? Annotations can only be on the parent object,
        # how to get the try_construct/external flags here?
        return xt.CommonCollectionElement(
            element_flags=flags,
            type=cls._xt_type_identifier("anonymous", _type, minimal)
        )

    @classmethod
    def _xt_minimal_array_type(cls, arr: pt.array) -> xt.MinimalArrayType:
        inner = arr
        while isinstance(inner, pt.array):
            inner = inner.subtype

        return xt.MinimalArrayType(
            collection_flag=xt.TypeFlag(),  # Unused, no flags apply
            header=cls._xt_minimal_array_header(arr),
            element=cls._xt_minimal_collection_element(inner, xt.TypeFlag(TRY_CONSTRUCT1=True))

        )

    @classmethod
    def _xt_complete_array_type(cls, arr: pt.array) -> xt.CompleteArrayType:
        inner = arr
        while isinstance(inner, pt.array):
            inner = inner.subtype

        return xt.CompleteArrayType(
            collection_flag=xt.TypeFlag(),  # Unused, no flags apply
            header=cls._xt_complete_array_header(arr),
            element=cls._xt_complete_collection_element(inner, xt.TypeFlag(TRY_CONSTRUCT1=True))

        )

    @classmethod
    def _xt_minimal_array_header(cls, arr: pt.array) -> xt.MinimalArrayHeader:
        return xt.MinimalArrayHeader(
            common=cls._xt_common_array_header(arr)
        )

    @classmethod
    def _xt_complete_array_header(cls, arr: pt.array) -> xt.CompleteArrayHeader:
        return xt.CompleteArrayHeader(
            common=cls._xt_common_array_header(arr),
            detail=cls._xt_complete_type_detail(arr)
        )

    @classmethod
    def _xt_common_array_header(cls, arr: pt.array) -> xt.CommonArrayHeader:
        inner = arr
        bounds = []
        while isinstance(inner, pt.array):
            bounds.append(inner.length)
            inner = inner.subtype

        return xt.CommonArrayHeader(
            bound_seq=bounds
        )

    @classmethod
    def _xt_minimal_bitmask_type(cls, entity: Type[IdlBitmask]) -> xt.MinimalBitmaskType:
        return xt.MinimalBitmaskType(
            bitmask_flags=cls._xt_type_flag(entity),
            header=cls._xt_minimal_enumerated_header(entity),
            flag_seq=cls._xt_minimal_bitflag_seq(entity)
        )

    @classmethod
    def _xt_complete_bitmask_type(cls, entity: Type[IdlBitmask]) -> xt.CompleteBitmaskType:
        return xt.CompleteBitmaskType(
            bitmask_flags=cls._xt_type_flag(entity),
            header=cls._xt_complete_enumerated_header(entity),
            flag_seq=cls._xt_complete_bitflag_seq(entity)
        )

    @classmethod
    def _xt_minimal_enumerated_header(cls, entity: Union[Type[IdlEnum], Type[IdlBitmask]]) -> xt.MinimalEnumeratedHeader:
        return xt.MinimalEnumeratedHeader(
            common=cls._xt_common_enumerated_header(entity)
        )

    @classmethod
    def _xt_complete_enumerated_header(cls, entity: Union[Type[IdlEnum], Type[IdlBitmask]]) -> xt.CompleteEnumeratedHeader:
        return xt.CompleteEnumeratedHeader(
            common=cls._xt_common_enumerated_header(entity),
            detail=cls._xt_complete_type_detail(entity)
        )

    @classmethod
    def _xt_common_enumerated_header(cls, entity: Union[Type[IdlEnum], Type[IdlBitmask]]) -> xt.CommonEnumeratedHeader:
        return xt.CommonEnumeratedHeader(
            bit_bound=get_idl_annotations(entity).get("bit_bound", 32)
        )

    @classmethod
    def _xt_minimal_bitflag_seq(cls, entity: Type[IdlBitmask]) -> xt.MinimalBitflagSeq:
        seq = [
            cls._xt_minimal_bitflag(entity, name)
            for name, _ in get_extended_type_hints(entity).items()
        ]
        return list(sorted(seq, key=lambda x: x.common.position))

    @classmethod
    def _xt_complete_bitflag_seq(cls, entity: Type[IdlBitmask]) -> xt.CompleteBitflagSeq:
        seq = [
            cls._xt_complete_bitflag(entity, name)
            for name, _ in get_extended_type_hints(entity).items()
        ]
        return list(sorted(seq, key=lambda x: x.common.position))

    @classmethod
    def _xt_minimal_bitflag(cls, entity: Type[IdlBitmask], name: str) -> xt.MinimalBitflag:
        return xt.MinimalBitflag(
            common=cls._xt_common_bit_flag(entity, name),
            detail=cls._xt_minimal_member_detail(entity, name, bool)
        )

    @classmethod
    def _xt_complete_bitflag(cls, entity: Type[IdlBitmask], name: str) -> xt.CompleteBitflag:
        return xt.CompleteBitflag(
            common=cls._xt_common_bit_flag(entity, name),
            detail=cls._xt_complete_member_detail(entity, name, bool)
        )

    @classmethod
    def _xt_common_bit_flag(cls, entity: Type[IdlBitmask], name: str) -> xt.CommonBitflag:
        return xt.CommonBitflag(
            position=entity.__idl_positions__[name],
            flags=xt.MemberFlag()  # Unused, no flags apply
        )

    @classmethod
    def _xt_complete_enumerated_type(cls, entity: Type[IdlEnum]) -> xt.CompleteEnumeratedType:
        return xt.CompleteEnumeratedType(
            enum_flags=cls._xt_type_flag(entity),
            header=cls._xt_complete_enumerated_header(entity),
            literal_seq=cls._xt_complete_enumerated_literal_seq(entity)
        )

    @classmethod
    def _xt_minimal_enumerated_type(cls, entity: Type[IdlEnum]) -> xt.MinimalEnumeratedType:
        return xt.MinimalEnumeratedType(
            enum_flags=cls._xt_type_flag(entity),
            header=cls._xt_minimal_enumerated_header(entity),
            literal_seq=cls._xt_minimal_enumerated_literal_seq(entity)
        )

    @classmethod
    def _xt_complete_enumerated_literal_seq(cls, entity: Type[IdlEnum]) -> xt.CompleteEnumeratedLiteralSeq:
        seq = [
            cls._xt_complete_enumerated_literal(entity, entry)
            for entry in entity
        ]
        return list(sorted(seq, key=lambda x: x.common.value))

    @classmethod
    def _xt_minimal_enumerated_literal_seq(cls, entity: Type[IdlEnum]) -> xt.MinimalEnumeratedLiteralSeq:
        seq = [
            cls._xt_minimal_enumerated_literal(entity, entry)
            for entry in entity
        ]
        return list(sorted(seq, key=lambda x: x.common.value))

    @classmethod
    def _xt_complete_enumerated_literal(cls, entity: Type[IdlEnum], instance: IdlEnum) -> xt.CompleteEnumeratedLiteral:
        return xt.CompleteEnumeratedLiteral(
            common=cls._xt_common_enumerated_literal(entity, instance),
            detail=cls._xt_complete_member_detail(entity, instance.name, entity)
        )

    @classmethod
    def _xt_minimal_enumerated_literal(cls, entity: Type[IdlEnum], instance: IdlEnum) -> xt.MinimalEnumeratedLiteral:
        return xt.MinimalEnumeratedLiteral(
            common=cls._xt_common_enumerated_literal(entity, instance),
            detail=cls._xt_minimal_member_detail(entity, instance.name, entity)
        )

    @classmethod
    def _xt_common_enumerated_literal(cls, entity: Type[IdlEnum], instance: IdlEnum) -> xt.CommonEnumeratedLiteral:
        return xt.CommonEnumeratedLiteral(
            value=instance.value,
            flags=xt.MemberFlag(IS_DEFAULT=instance.value == entity.__idl_enum_default_value__)
        )


class XTParseState:
    class Deferred:
        def __init__(self):
            self.callbacks = []

    def __init__(self, database):
        self.database: Dict[xt.TypeIdentifier, xt.TypeObject] = database
        self.resolved: Dict[xt.TypeIdentifier, Any] = {}
        self.reentrant: Set[xt.TypeIdentifier] = set()
        self.waiting: Dict[xt.TypeIdentifier, List[XTParseState.DeferredMember]] = {}


class XTInterpreter:
    _simple_mappings = {
        xt.TK_BOOLEAN: bool,
        xt.TK_BYTE: pt.byte,
        xt.TK_INT8: pt.int8,
        xt.TK_INT16: pt.int16,
        xt.TK_INT32: pt.int32,
        xt.TK_INT64: pt.int64,
        xt.TK_UINT8: pt.uint8,
        xt.TK_UINT16: pt.uint16,
        xt.TK_UINT32: pt.uint32,
        xt.TK_UINT64: pt.uint64,
        xt.TK_FLOAT32: pt.float32,
        xt.TK_FLOAT64: pt.float64,
        # float128 omitted
        xt.TK_CHAR8: pt.char,
        xt.TK_CHAR16: pt.wchar,
        xt.TK_STRING8: str
    }

    @classmethod
    def xt_to_class(cls, ident: xt.TypeIdentifier,
                    tmap: Dict[xt.TypeIdentifier, xt.TypeObject]) -> \
            Tuple[Union[IdlUnion, IdlStruct], Dict[str, Union[IdlUnion, IdlStruct, IdlEnum, IdlBitmask]]]:
        state = XTParseState(tmap)

        main_type = cls._from_typeid(ident, state)
        all_types = {
            (c.name if isinstance(c, pt.typedef) else c.__name__): c
            for c in state.resolved.values()
        }
        return main_type, all_types

    @classmethod
    def _from_typeid(cls, ident: xt.TypeIdentifier, state: XTParseState) -> Any:
        tk = ident.discriminator

        if not tk or tk == xt.TK_NONE:
            return None
        elif tk in cls._simple_mappings:
            return cls._simple_mappings[tk]
        elif tk in [xt.TK_FLOAT128, xt.TK_STRING16, xt.TI_STRING16_SMALL, xt.TI_STRING16_LARGE]:
            raise NotImplementedError("Float128 and wstring types are not supported in Python.")
        elif tk in [xt.TI_STRING8_SMALL, xt.TI_STRING8_LARGE]:
            descriptor: Union[xt.StringSTypeDefn, xt.StringLTypeDefn] = ident.value
            if descriptor.bound == 0 or descriptor.bound == uint32_max:
                return str
            return pt.bounded_str[descriptor.bound]
        elif tk in [xt.TI_PLAIN_SEQUENCE_SMALL, xt.TI_PLAIN_SEQUENCE_LARGE]:
            descriptor: Union[xt.PlainSequenceSElemDefn, xt.PlainSequenceLElemDefn] = ident.value
            if descriptor.header.element_flags.IS_OPTIONAL:
                if descriptor.bound:
                    return Optional[pt.sequence[cls._from_typeid(descriptor.element_identifier, state), descriptor.bound]]
                return Optional[pt.sequence[cls._from_typeid(descriptor.element_identifier, state)]]

            if descriptor.bound:
                return pt.sequence[cls._from_typeid(descriptor.element_identifier, state), descriptor.bound]
            return pt.sequence[cls._from_typeid(descriptor.element_identifier, state)]
        elif tk in [xt.TI_PLAIN_ARRAY_SMALL, xt.TI_PLAIN_ARRAY_LARGE]:
            descriptor: Union[xt.PlainArraySElemDefn, xt.PlainArrayLElemDefn] = ident.value
            inner = cls._from_typeid(descriptor.element_identifier, state)

            for bound in reversed(descriptor.array_bound_seq):
                inner = pt.array[inner, bound]

            if descriptor.header.element_flags.IS_OPTIONAL:
                inner = Optional[inner]

            return inner
        # add map types...
        else:
            return cls._from_typeobject(ident, state)

    @classmethod
    def _from_typeobject(cls, ident: xt.TypeIdentifier, state: XTParseState) -> Any:
        if ident in state.resolved:
            return state.resolved[ident]

        if ident in state.reentrant:
            defer = XTParseState.Deferred()
            state.waiting[ident].append(defer)
            return defer

        state.reentrant.add(ident)
        state.waiting[ident] = []

        obj = state.database[ident]
        if (obj.discriminator & xt.EK_COMPLETE) == xt.EK_COMPLETE:
            complete_obj: xt.CompleteTypeObject = obj.complete
            tk = complete_obj.discriminator

            if tk == xt.TK_STRUCTURE:
                return cls._make_complete_struct(ident, complete_obj.struct_type, state)
            elif tk == xt.TK_UNION:
                return cls._make_complete_union(ident, complete_obj.union_type, state)
            elif tk == xt.TK_BITMASK:
                return cls._make_complete_bitmask(ident, complete_obj.bitmask_type, state)
            elif tk == xt.TK_ARRAY:
                return cls._make_complete_array(ident, complete_obj.array_type, state)
            elif tk == xt.TK_SEQUENCE:
                return cls._make_complete_sequence(ident, complete_obj.sequence_type, state)
            elif tk == xt.TK_ENUM:
                return cls._make_complete_enum(ident, complete_obj.enumerated_type, state)
            elif tk == xt.TK_ALIAS:
                return cls._make_complete_alias(ident, complete_obj.alias_type, state)
            # map types, annotation types
            raise NotImplementedError()
        else:
            # We don't convert minimal type ids back into types
            raise NotImplementedError()

    @classmethod
    def _resolve_reentrant(cls, ident: xt.TypeIdentifier, resolver: Union[IdlStruct, IdlUnion], state: XTParseState) -> None:
        state.reentrant.remove(ident)
        state.resolved[ident] = resolver

        for waiter in state.waiting[ident]:
            for callback in waiter.callbacks:
                callback()

        del state.waiting[ident]

    @classmethod
    def _make_complete_struct(cls, ident: xt.TypeIdentifier, pre_struct: xt.CompleteStructType,
                              state: XTParseState) -> Type[IdlStruct]:
        base = cls._from_typeid(pre_struct.header.base_type, state)

        types = {}
        annotations = {}

        if base is None:
            bases = tuple()
            member_ids = {}
        else:
            bases = (base,)
            member_ids = base.__idl__.member_ids
        defers = []

        for m in pre_struct.member_seq:
            m_id = m.common.member_id
            m_type = cls._from_typeid(m.common.member_type_id, state)
            m_name = m.detail.name
            if m_name in python_keywords:
                m_name = "_" + m_name
            m_annotations = {}

            if m.common.member_flags.IS_MUST_UNDERSTAND:
                m_annotations["must_understand"] = True

            if m.common.member_flags.IS_KEY:
                m_annotations["key"] = True

            if m.common.member_flags.IS_EXTERNAL:
                m_annotations["external"] = True

            if m_name != m.detail.name:
                m_annotations["name"] = m.detail.name

            if isinstance(m_type, XTParseState.Deferred):
                defers.append((m_name, m.common.member_type_id, m_type, m.common.member_flags.IS_OPTIONAL))
                m_type = Any
            elif m.common.member_flags.IS_OPTIONAL:
                m_type = Optional[m_type]

            types[m_name] = m_type
            if m_annotations:
                annotations[m_name] = m_annotations
            member_ids[m_name] = m_id

        struct = make_idl_struct(
            class_name=pre_struct.header.detail.type_name.split('::')[-1],
            typename=pre_struct.header.detail.type_name.replace('::', '.'),
            fields=types,
            bases=bases,
            field_annotations=annotations
        )
        struct.__idl__.member_ids = member_ids

        if pre_struct.struct_flags.IS_FINAL:
            struct = annotate.final(struct)
        if pre_struct.struct_flags.IS_APPENDABLE:
            struct = annotate.appendable(struct)
        if pre_struct.struct_flags.IS_MUTABLE:
            struct = annotate.mutable(struct)
        if pre_struct.struct_flags.IS_NESTED:
            struct = annotate.nested(struct)
        if pre_struct.struct_flags.IS_AUTOID_HASH:
            struct = annotate.autoid("hash")(struct)

        for (m_name, type_ident, defer, is_optional) in defers:
            def def_callback():
                _type = XTInterpreter._from_typeid(type_ident, state)
                struct.__annotations__[m_name] = Optional[_type] if is_optional else _type

            defer.callbacks.append(def_callback)

        cls._resolve_reentrant(ident, struct, state)
        return struct

    @classmethod
    def _make_complete_union(cls, ident: xt.TypeIdentifier, pre_union: xt.CompleteUnionType,
                             state: XTParseState) -> Type[IdlStruct]:
        discriminator = cls._from_typeid(pre_union.discriminator.common.type_id, state)

        cases = {}
        annotations = {}
        member_ids = {}
        defers = []

        for m in pre_union.member_seq:
            m_id = m.common.member_id
            m_type = cls._from_typeid(m.common.type_id, state)
            m_name = m.detail.name
            if m_name in ['value', 'discriminator'] or m_name in python_keywords:
                m_name = "_" + m_name
            m_annotations = {}

            if m.common.member_flags.IS_MUST_UNDERSTAND:
                m_annotations["must_understand"] = True

            if m.common.member_flags.IS_EXTERNAL:
                m_annotations["external"] = True

            if m_name != m.detail.name:
                m_annotations["name"] = m.detail.name

            if isinstance(m_type, XTParseState.Deferred):
                defers.append((m_name, m.common.member_type_id, m_type))
                m_type = Any

            if m.common.member_flags.IS_DEFAULT or not m.common.label_seq:
                m_type = pt.default[m_type]
            else:
                if isclass(discriminator) and issubclass(discriminator, IdlEnum):
                    m_type = pt.case[[discriminator(l) for l in m.common.label_seq], m_type]
                else:
                    m_type = pt.case[m.common.label_seq, m_type]

            cases[m_name] = m_type
            if m_annotations:
                annotations[m_name] = m_annotations
            member_ids[m_name] = m_id

        union = make_idl_union(
            class_name=pre_union.header.detail.type_name.split('::')[-1],
            typename=pre_union.header.detail.type_name.replace('::', '.'),
            fields=cases,
            field_annotations=annotations,
            discriminator=discriminator,
            discriminator_is_key=pre_union.discriminator.common.member_flags.IS_KEY
        )
        union.__idl__.member_ids = member_ids

        if pre_union.union_flags.IS_FINAL:
            union = annotate.final(union)
        if pre_union.union_flags.IS_APPENDABLE:
            union = annotate.appendable(union)
        if pre_union.union_flags.IS_MUTABLE:
            union = annotate.mutable(union)
        if pre_union.union_flags.IS_NESTED:
            union = annotate.nested(union)
        if pre_union.union_flags.IS_AUTOID_HASH:
            union = annotate.autoid("hash")(union)

        for (m_name, type_ident, defer) in defers:
            def def_callback():
                _type = XTInterpreter._from_typeid(type_ident, state)
                union.__annotations__[m_name] = _type

            defer.callbacks.append(def_callback)

        cls._resolve_reentrant(ident, union, state)
        return union

    @classmethod
    def _make_complete_bitmask(cls, ident: xt.TypeIdentifier, bitmask_type: xt.CompleteBitmaskType,
                               state: XTParseState) -> Type[IdlStruct]:
        bitmask = make_idl_bitmask(
            class_name=bitmask_type.header.detail.type_name.split('::')[-1],
            typename=bitmask_type.header.detail.type_name.replace('::', '.'),
            fields=[flag.detail.name for flag in bitmask_type.flag_seq],
            dataclassify=True,
            field_annotations={
                flag.detail.name: {
                    'position': flag.common.position
                }
                for flag in bitmask_type.flag_seq
            }
        )

        if bitmask_type.header.common.bit_bound != 32:
            annotate.bit_bound(bitmask_type.header.common.bit_bound)(bitmask)

        if bitmask_type.bitmask_flags.IS_FINAL:
            bitmask = annotate.final(bitmask)
        if bitmask_type.bitmask_flags.IS_APPENDABLE:
            bitmask = annotate.appendable(bitmask)
        if bitmask_type.bitmask_flags.IS_MUTABLE:
            bitmask = annotate.mutable(bitmask)

        cls._resolve_reentrant(ident, bitmask, state)
        return bitmask

    @classmethod
    def _make_complete_enum(cls, ident: xt.TypeIdentifier, enum_type: xt.CompleteEnumeratedType,
                            state: XTParseState) -> Type[IdlStruct]:
        # Enums cannot be re-entrant, no deps so no possible cycles.

        _default = None
        for literal in enum_type.literal_seq:
            if literal.common.flags.IS_DEFAULT:
                _default = literal.detail.name
                break

        enum_ = make_idl_enum(
            class_name=enum_type.header.detail.type_name.split('::')[-1],
            typename=enum_type.header.detail.type_name.replace('::', '.'),
            fields={
                literal.detail.name: literal.common.value
                for literal in enum_type.literal_seq
            },
            default=_default
        )

        if enum_type.enum_flags.IS_FINAL:
            enum_ = annotate.final(enum_)
        if enum_type.enum_flags.IS_APPENDABLE:
            enum_ = annotate.appendable(enum_)
        if enum_type.enum_flags.IS_MUTABLE:
            enum_ = annotate.mutable(enum_)

        cls._resolve_reentrant(ident, enum_, state)
        return enum_

    @classmethod
    def _make_complete_sequence(cls, ident: xt.TypeIdentifier, pre_seq: xt.CompleteSequenceType,
                                state: XTParseState) -> pt.sequence:
        inner = cls._from_typeid(pre_seq.element.common.type, state)

        if isinstance(inner, XTParseState.Deferred):
            return inner

        if pre_seq.header.common.bound:
            return pt.sequence[inner, pre_seq.header.common.bound]

        seq = pt.sequence[inner]
        cls._resolve_reentrant(ident, seq, state)
        return seq

    @classmethod
    def _make_complete_array(cls, ident: xt.TypeIdentifier, pre_arr: xt.CompleteArrayType, state: XTParseState) -> pt.array:
        inner = cls._from_typeid(pre_arr.element.common.type, state)

        if isinstance(inner, XTParseState.Deferred):
            return inner

        for bound in reversed(pre_arr.header.common.bound_seq):
            inner = pt.array[inner, bound]

        cls._resolve_reentrant(ident, inner, state)
        return inner

    @classmethod
    def _make_complete_alias(cls, ident: xt.TypeIdentifier, pre_alias: xt.CompleteAliasType,
                             state: XTParseState) -> pt.typedef:
        inner = cls._from_typeid(pre_alias.body.common.related_type, state)

        if isinstance(inner, XTParseState.Deferred):
            return inner

        typedef = pt.typedef(pre_alias.header.detail.type_name, inner)
        cls._resolve_reentrant(ident, typedef, state)
        return typedef


class XTTypeIdScanner:
    """Utility to discover all Type Id's mentioned in a TypeObject"""
    @classmethod
    def find_all_typeids(cls, type_object: xt.TypeObject) -> List[xt.TypeIdentifier]:
        if (type_object.discriminator & xt.EK_COMPLETE) == xt.EK_COMPLETE:
            return cls._scan_complete(type_object.complete)
        raise NotImplementedError()

    @classmethod
    def _scan_complete(cls, complete_type_object: xt.CompleteTypeObject) -> List[xt.TypeIdentifier]:
        tk = complete_type_object.discriminator

        if tk == xt.TK_STRUCTURE:
            return cls._scan_complete_struct(complete_type_object.struct_type)
        elif tk == xt.TK_UNION:
            return cls._scan_complete_union(complete_type_object.union_type)
        elif tk == xt.TK_ARRAY:
            return cls._scan_complete_array(complete_type_object.array_type)
        elif tk == xt.TK_SEQUENCE:
            return cls._scan_complete_sequence(complete_type_object.sequence_type)
        elif tk == xt.TK_ALIAS:
            return cls._scan_complete_alias(complete_type_object.alias_type)
        elif tk in [xt.TK_ENUM, xt.TK_BITMASK]:
            return []  # No deps
        # Unsupported: map, annotation, bitset
        raise NotImplementedError()

    @classmethod
    def _scan_complete_struct(cls, complete_struct: xt.CompleteStructType):
        ret = []
        if complete_struct.header.base_type.discriminator != xt.TK_NONE:
            ret.append(complete_struct.header.base_type)

        for member in complete_struct.member_seq:
            # EK_BOTH signifies _plain_
            if member.common.member_type_id.discriminator != xt.EK_BOTH:
                ret += cls._scan_type_id(member.common.member_type_id)

        return ret

    @classmethod
    def _scan_complete_union(cls, complete_union: xt.CompleteUnionType):
        ret = []

        if complete_union.discriminator.common.type_id.discriminator != xt.EK_BOTH:
            ret += cls._scan_type_id(complete_union.discriminator.common.type_id)

        for member in complete_union.member_seq:
            # EK_BOTH signifies _plain_
            if member.common.type_id.discriminator != xt.EK_BOTH:
                ret += cls._scan_type_id(member.common.type_id)

        return ret

    @classmethod
    def _scan_complete_sequence(cls, pre_seq: xt.CompleteSequenceType):
        return cls._scan_type_id(pre_seq.element.common.type)

    @classmethod
    def _scan_complete_array(cls, pre_arr: xt.CompleteArrayType):
        return cls._scan_type_id(pre_arr.element.common.type)

    @classmethod
    def _scan_complete_alias(cls, pre_alias: xt.CompleteAliasType):
        return cls._scan_type_id(pre_alias.body.common.related_type)

    @classmethod
    def _scan_type_id(cls, type_id: xt.TypeIdentifier):
        tk = type_id.discriminator

        if tk <= xt.TK_CHAR16:
            # simple type
            return []
        if tk in [xt.TI_STRING8_SMALL, xt.TI_STRING16_SMALL, xt.TI_STRING8_LARGE,
                  xt.TI_STRING16_LARGE, xt.TK_STRING8, xt.TK_STRING16]:
            return []
        elif tk in [xt.TI_PLAIN_SEQUENCE_SMALL, xt.TI_PLAIN_SEQUENCE_LARGE]:
            descriptor: Union[xt.PlainSequenceSElemDefn, xt.PlainSequenceLElemDefn] = type_id.value
            return cls._scan_type_id(descriptor.element_identifier)
        elif tk in [xt.TI_PLAIN_ARRAY_SMALL, xt.TI_PLAIN_ARRAY_LARGE]:
            descriptor: Union[xt.PlainArraySElemDefn, xt.PlainArrayLElemDefn] = type_id.value
            return cls._scan_type_id(descriptor.element_identifier)
        return [type_id]



"""
 * Copyright(c) 2021 to 2022 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""

from sys import exit
from pathlib import Path


idlpy_path = list(Path(__file__).resolve().parent.glob("_idlpy*"))[0]


if __name__ == "__main__":
    print(idlpy_path, end="", flush=True)
    exit(0)



import re
from datetime import datetime, timedelta
from rich.layout import Layout
from rich.text import Text

from .app import Header, CPUGraph, ScrollGraph, MultiScrollGraph, PeerPanel
from .barchart import RichChart


def make_ping_layout(cmd) -> Layout:
    layout = Layout(name="ddsperf_app")

    layout.split(Layout(name="header", size=3), Layout(name="main", ratio=1))
    layout["main"].split_row(
        Layout(name="latencies_perhost", ratio=1), Layout(name="others", ratio=1)
    )
    layout["others"].split_column(
        Layout(name="top_others", ratio=1), Layout(name="bottom_others", ratio=1)
    )
    layout["top_others"].split_row(
        Layout(name="latencies_combined", ratio=2), Layout(name="peers", ratio=1)
    )
    layout["bottom_others"].split_row(
        Layout(name="mini1", ratio=1),
        Layout(name="mini2", ratio=1),
        Layout(name="mini3", ratio=1),
    )
    layout["mini1"].split_column(Layout(name="maxrss"), Layout(name="discarded"))
    layout["mini2"].split_column(Layout(name="rexmit"), Layout(name="trexmit"))
    layout["mini3"].split_column(Layout(name="tthrottle"), Layout(name="nthrottle"))

    layout["header"].update(Header(" ".join(cmd)))
    layout["latencies_perhost"].update(Text(""))
    layout["latencies_combined"].update(Text(""))
    layout["peers"].update(Text(""))
    layout["maxrss"].update(Text(""))
    layout["discarded"].update(Text(""))
    layout["rexmit"].update(Text(""))
    layout["trexmit"].update(Text(""))
    layout["tthrottle"].update(Text(""))
    layout["nthrottle"].update(Text(""))
    return layout


def make_ping_updater():
    data_re = re.compile(
        r"\[\d*\] [\d\.]+\s+([^\s:]+):(\d+) size \d+ mean ([\d\.]+)(n|u|m|s)s min ([\d\.]+)(n|u|m|s)s 50% ([\d\.]+)(n|u|m|s)s 90% ([\d\.]+)(n|u|m|s)s 99% ([\d\.]+)(n|u|m|s)s max (?:[\d\.]+)(?:n|u|m|s)s cnt (\d+)"
    )
    peer_re = re.compile(r"\[\d+\] participant ([^:]+):(\d+): (.*)$")
    rss_re = re.compile(r"rss:([\d\.]+)(kb|MB)")
    discarded_re = re.compile(r"discarded ([\d\.]+)")
    rexmit_re = re.compile(r" rexmit ([\d\.]+)")
    trexmit_re = re.compile(r" Trexmit ([\d\.]+)")
    tthrottle_re = re.compile(r" Tthrottle ([\d\.]+)")
    nthrottle_re = re.compile(r" Nthrottle ([\d\.]+)")

    cols = ["green", "blue", "magenta", "cyan", "white", "red"]

    graphs = {}
    graphs["peers"] = PeerPanel(True)
    graphs["latencies_perhost"] = ScrollGraph(
        "Latency per host (waiting for data)", 10, "bright_cyan"
    )
    graphs["latencies_combined"] = MultiScrollGraph("Mean Ping latency (µs)", 20)
    graphs["maxrss"] = ScrollGraph("MaxRSS", 10, "bright_cyan")
    graphs["discarded"] = ScrollGraph("Discarded msg", 10, "bright_green", delta=True)
    graphs["rexmit"] = ScrollGraph("Rexmit", 10, "bright_green", delta=True)
    graphs["trexmit"] = ScrollGraph("TRexmit", 10, "bright_green")
    graphs["tthrottle"] = ScrollGraph("Tthrottle", 10, "bright_green")
    graphs["nthrottle"] = ScrollGraph("Nthrottle", 10, "bright_green", delta=True)

    latency_graphs = []
    host_index = {}
    categories = ["mean", "min", "50%", "90%", "99%"]  # , "max"]
    data = [0] * len(categories)
    mult = {"n": 0.001, "u": 1, "m": 1000, "s": 1000000}

    counter = 0
    timepoint = datetime.now()

    def updater(line):
        nonlocal timepoint, counter
        updated = set()

        if datetime.now() - timepoint > timedelta(seconds=3):
            timepoint = datetime.now()
            if len(latency_graphs) > 1:
                counter = (counter + 1) % len(latency_graphs)
                updated.add("latencies_perhost")
                latency_graphs[counter].update_chart_vars(
                    **graphs["latencies_perhost"]._chart_kwargs
                )
                graphs["latencies_perhost"] = latency_graphs[counter]

        m = data_re.match(line)
        if m:
            updated.add("latencies_perhost")
            updated.add("latencies_combined")
            host, pid = m.group(1, 2)
            for i in range(len(categories)):
                data[i] = float(m.group(3 + 2 * i)) * mult[m.group(4 + 2 * i)]
            # count = m.group(15)

            key = f"{host}:{pid}"
            if key not in host_index:
                color = graphs["peers"].get_peer_color(host, pid)
                index = graphs["latencies_combined"].add_new_line(color)
                host_index[key] = index

                counter = len(latency_graphs)
                latency_graphs.append(
                    MultiScrollGraph(f"[bold][{color}]{host}[/]:{pid}[/]", 20)
                )
                latency_graphs[-1].update_chart_vars(
                    **graphs["latencies_perhost"]._chart_kwargs
                )
                graphs["latencies_perhost"] = latency_graphs[-1]
                for i in range(len(categories)):
                    latency_graphs[-1].add_new_line(cols[i])

            index = host_index[key]
            for i in range(len(categories)):
                latency_graphs[index].add_point(i, data[i])
            graphs["latencies_combined"].add_point(index, data[0])

        rss = rss_re.search(line)
        if rss:
            updated.add("maxrss")
            graphs["maxrss"].add_point(
                float(rss.group(1)) * (1000000 if rss.group(2) == "MB" else 1000)
            )

        discarded_m = discarded_re.search(line)
        if discarded_m:
            updated.add("discarded")
            graphs["discarded"].add_point(int(discarded_m.group(1)))

        rexmit_m = rexmit_re.search(line)
        if rexmit_m:
            updated.add("rexmit")
            graphs["rexmit"].add_point(int(rexmit_m.group(1)))

        trexmit_m = trexmit_re.search(line)
        if trexmit_m:
            updated.add("trexmit")
            graphs["trexmit"].add_point(int(trexmit_m.group(1)))

        tthrottle_m = tthrottle_re.search(line)
        if tthrottle_m:
            updated.add("tthrottle")
            graphs["tthrottle"].add_point(int(tthrottle_m.group(1)))

        nthrottle_m = nthrottle_re.search(line)
        if nthrottle_m:
            updated.add("nthrottle")
            graphs["nthrottle"].add_point(int(nthrottle_m.group(1)))

        peer_m = peer_re.match(line)
        if peer_m:
            updated.add("peers")
            host, pid, action = peer_m.group(1), peer_m.group(2), peer_m.group(3)
            if action == "new (self)":
                graphs["peers"].set_own(host, pid)
            elif action == "new":
                graphs["peers"].add_peer(host, pid)
            elif action == "gone":
                graphs["peers"].remove_peer(host, pid)

        return updated

    return updater, graphs



"""
 * Copyright(c) 2021 to 2022 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""

import sys
import json
import datetime
import argparse

from cyclonedds.domain import DomainParticipant
from cyclonedds.builtin import (BuiltinDataReader, BuiltinTopicDcpsParticipant,
                                BuiltinTopicDcpsSubscription, BuiltinTopicDcpsPublication)
from cyclonedds.util import duration
from cyclonedds.core import WaitSet, ReadCondition, ViewState, InstanceState, SampleState


class TopicManager:
    def __init__(self, reader, topic_type, args):
        self.reader = reader
        self.topic_type = topic_type
        self.console_print = not args.filename
        self.enable_json = args.json
        self.enable_view = args.verbose
        self.dp_key = reader.participant.guid
        self.tracked_entities = {}
        self.tracked_disposed_entities = {}
        self.qoses = {}
        self.strings = ""
        self.read_cond = ReadCondition(reader, ViewState.Any | InstanceState.Alive | SampleState.NotRead)
        self.disposed_cond = ReadCondition(reader, ViewState.Any | InstanceState.NotAliveDisposed | SampleState.Any)

    # Read and print samples
    def poll(self):
        new_samples = []
        newly_disposed_samples = []

        while True:
            samples = self.reader.take(N=100, condition=self.read_cond)

            if not samples:
                break

            for sample in samples:
                if (
                    (self.topic_type == "PARTICIPANT" and sample.key != self.dp_key)
                    or (self.topic_type != "PARTICIPANT" and sample.participant_key != self.dp_key)
                ):
                    self.track_sample(sample)
                    if self.check_qos_changes(sample):
                        new_samples.append(sample)

        while True:
            disposed_samples = self.reader.take(N=100, condition=self.disposed_cond)

            if not disposed_samples:
                break

            for disposed_sample in disposed_samples:
                newly_disposed_samples.append(self.tracked_entities.get(str(disposed_sample.key), disposed_sample))
                self.untrack_sample(disposed_sample)

        if self.console_print and (new_samples or newly_disposed_samples):
            self.to_console(new_samples, newly_disposed_samples)

    # Add new samples to dict
    def track_sample(self, sample):
        self.tracked_entities[str(sample.key)] = sample

    # Collect disposed sample to a new dict and delete it from the original dict
    def untrack_sample(self, sample):
        if str(sample.key) in self.tracked_entities:
            self.tracked_disposed_entities[str(sample.key)] = self.tracked_entities[str(sample.key)]
            del self.tracked_entities[str(sample.key)]

    # Track qos changes
    def check_qos_changes(self, sample):
        key = sample.key
        if self.qoses.get(key, 0) == 0:
            self.qoses[key] = sample.qos
        elif self.qoses[key] != sample.qos:
            print(f"""\n\033[1mQos changed on topic '{sample.topic_name}' {self.topic_type.lower()}:\033[0m
                  \r key = {sample.key}""")
            for i in self.qoses[key]:
                if self.qoses[key][i] != sample.qos[i]:
                    print(f"\033[1m {str(self.qoses[key][i])} -> {str(sample.qos[i])}\033[0m")
            self.qoses[key] = sample.qos

            if not self.enable_view:  # If not verbose, don't print the qos changed sample data
                sample = None
        return sample

    # Wait for new / disposed samples
    def add_to_waitset(self, waitset):
        waitset.attach(self.read_cond)
        waitset.attach(self.disposed_cond)

    # Format sample value for console printing in JSON
    def format_value(self, sample):
        if self.topic_type == "PARTICIPANT":
            return {
                "key": str(sample.key),
            }
        else:
            return {
                "key": str(sample.key),
                "participant_key": str(sample.participant_key),
                "topic_name": str(sample.topic_name),
                "type_name": str(sample.type_name),
                "qos": sample.qos.asdict()
            }

    # Format sample for writing to file in JSON
    def as_dict(self):
        return {
            "New": {k: self.format_value(v) for k, v in self.tracked_entities.items()},
            "Disposed": {k: self.format_value(v) for k, v in self.tracked_disposed_entities.items()}
        }

    # Print to console in JSON or non-JSON format
    def to_console(self, new_samples, newly_disposed_samples):
        if self.enable_json:
            if new_samples:
                JsonWriter.write({
                    "type": self.topic_type,
                    "event": "new",
                    "value": [self.format_value(sample) for sample in new_samples]
                })
            if newly_disposed_samples:
                JsonWriter.write({
                    "type": self.topic_type,
                    "event": "disposed",
                    "value": [self.format_value(sample) for sample in newly_disposed_samples]
                })
        else:
            if new_samples:
                print(f"\n-- New {self.topic_type} --")
                for sample in new_samples:
                    print("\n".join(f" {item}: {sample.__dict__[item]}" for item in sample.__dict__
                          if item != "sample_info") + "\n")
            if newly_disposed_samples:
                print(f"\n-- Disposed {self.topic_type} --")
                for sample in newly_disposed_samples:
                    print("\n".join(f" {item}: {sample.__dict__[item]}" for item in sample.__dict__
                          if item != "sample_info") + "\n")


# Check the topic(s)
def parse_args(args):
    if args.topic:
        if args.topic == "dcpsparticipant":
            topic = [["PARTICIPANT", BuiltinTopicDcpsParticipant]]
        elif args.topic == "dcpssubscription":
            topic = [["SUBSCRIPTION", BuiltinTopicDcpsSubscription]]
        else:
            topic = [["PUBLICATION", BuiltinTopicDcpsPublication]]

    if args.all is True:
        topic = [["PARTICIPANT", BuiltinTopicDcpsParticipant],
                 ["SUBSCRIPTION", BuiltinTopicDcpsSubscription],
                 ["PUBLICATION", BuiltinTopicDcpsPublication]]
    return topic


def create_parser(args):
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--id", type=int, help="Define the domain participant id")
    parser.add_argument("-f", "--filename", type=str, help="Write results to file in JSON format")
    parser.add_argument("-j", "--json", action="store_true", help="Print output in JSON format")
    parser.add_argument("-w", "--watch", action="store_true", help="Watch for data reader & writer & qoses changes")
    parser.add_argument("-v", "--verbose", action="store_true",
                        help="View the sample when Qos changes")
    parser.add_argument("-r", "--runtime", type=float, help="Limit the runtime of the tool, in seconds.")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-a", "--all", action="store_true", help="for all topics")
    group.add_argument("-t", "--topic", choices=["dcpsparticipant", "dcpssubscription", "dcpspublication"],
                       help="for one specific topic")

    args = parser.parse_args(args)
    return args


# Format samples in JSON
class JsonWriter:
    first = True

    @classmethod
    def write(cls, data):
        if not cls.first:
            print(",")
        cls.first = False
        json.dump(data, sys.stdout, indent=4)

    @classmethod
    def reset(cls):
        cls.first = True


def main(sys_args):
    JsonWriter.reset()
    managers = []
    args = create_parser(sys_args)
    if args.id is None:
        dp = DomainParticipant()
    else:
        dp = DomainParticipant(args.id)
    topics = parse_args(args)
    waitset = WaitSet(dp)

    for topic_type, topic in topics:
        # Create TopicManager for each topic
        managers.append(TopicManager(BuiltinDataReader(dp, topic), topic_type, args))
        managers[-1].add_to_waitset(waitset)

    if not args.filename and args.json:
        print("[")

    # Watchmode
    if args.watch:
        try:
            time_start = datetime.datetime.now()
            v = True
            while v:
                for manager in managers:
                    waitset.wait(duration(milliseconds=20))
                    manager.poll()
                if args.runtime:
                    v = datetime.datetime.now() < time_start + datetime.timedelta(seconds=args.runtime)
        except KeyboardInterrupt:
            pass
    # Non-watchmode
    else:
        time_start = datetime.datetime.now()
        runtime = args.runtime or 1
        while datetime.datetime.now() < time_start + datetime.timedelta(seconds=runtime):
            for manager in managers:
                manager.poll()

    if not args.filename and args.json:
        print("]")

    # Write to file
    if args.filename:
        try:
            with open(args.filename, 'w') as f:
                data = {manager.topic_type: manager.as_dict() for manager in managers}
                json.dump(data, f, indent=4)
                print(f"\nResults have been written to file {args.filename}\n")
        except OSError:
            raise Exception(f"Could not open file {args.filename}")
    return 0


def command():
    main(sys.argv[1:])



"""
 * Copyright(c) 2021 to 2022 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""

import sys


if sys.version_info < (3, 9):
    raise NotImplementedError("This package cannot be used in Python version 3.8 or lower.")
else:
    # We are in any Python 3.9 or 3.10 (maybe higher?) version
    from typing import Annotated, get_origin, get_args, get_type_hints  # noqa F401


__all__ = ["Annotated", "get_origin", "get_args", "get_type_hints"]



import asyncio
import pytest

import cyclonedds.idl as idl
import cyclonedds.idl.annotations as annotate

from dataclasses import dataclass
from cyclonedds.domain import DomainParticipant
from cyclonedds.topic import Topic
from cyclonedds.pub import DataWriter
from cyclonedds.sub import DataReader
from cyclonedds.util import duration


@dataclass
@annotate.mutable
@annotate.autoid("sequential")
class Base(idl.IdlStruct, typename="Hierarchy.Base"):
    fieldA: str


@dataclass
@annotate.mutable
@annotate.autoid("sequential")
class Derived(Base, typename="Hierarchy.Derived"):
    fieldB: str


@pytest.mark.asyncio
async def test_base_topic():
    '''
    Creates a publisher and a subscriber of the Base topic and checks if the
    update sent by the publisher is received by the subscriber.
    '''
    base = Base(fieldA='Lorem')
    tasks = [
        _subscriber(Base),
        _publisher(Base, base),
    ]
    results = await asyncio.gather(*tasks)
    assert results[0] == results[1]


@pytest.mark.asyncio
async def test_derived_topic():
    '''
    Creates a publisher and a subscriber of the Derived topic and checks if the
    update sent by the publisher is received by the subscriber.
    '''
    derived = Derived(
        fieldA='Ipsum',
        fieldB='Dolor')

    tasks = [
        _subscriber(Derived),
        _publisher(Derived, derived),
    ]

    results = await asyncio.gather(*tasks)
    assert results[0] == results[1]


@pytest.mark.asyncio
async def test_base_and_derived_topics():
    '''
    Creates publishers and a subscribers of the Base and Derived topics and
    checks if the updates sent by the publishers are received by the
    subscribers.
    '''
    base = Base(fieldA='Lorem')
    derived = Derived(
        fieldA='Ipsum',
        fieldB='Dolor')

    tasks = [
        # This should not be necessary. I suspect there is a bug in CycloneDDS
        _subscriber(Base),
        _publisher(Base, base),

        # This is the part of the test that really matters
        _subscriber(Derived),
        _publisher(Derived, derived),
    ]

    results = await asyncio.gather(*tasks)
    assert results[0] == results[1]
    assert results[2] == results[3]


async def _publisher(topicClass, value, timeout=2):
    '''
    Publishes an update with a given value.
    '''
    participant = DomainParticipant(0)
    topic = Topic(participant, topicClass.__name__, topicClass)
    writer = DataWriter(participant, topic)
    writer.write(value)
    await asyncio.sleep(timeout)
    return value


async def _subscriber(topicClass, timeout=2):
    '''
    Receives an update. Raises an exception if no update is received within a
    given timeout.
    '''
    participant = DomainParticipant(0)
    topic = Topic(participant, topicClass.__name__, topicClass)
    reader = DataReader(participant, topic)
    async for update in reader.read_aiter(timeout=duration(seconds=timeout)):
        return update
